import logging
import numpy as np
from .shape import Circle, Arc, Line, Element
from .functions import distance

logger = logging.getLogger(__name__)

def polylines(entity, lf, rf, xoff=0.0, yoff=0.0, rotation=0.0):
    """returns a collection of bulged vertices
    http://www.afralisp.net/archive/lisp/Bulges1.htm
    """
    if isinstance(entity.points, list):
        points = [(p[0], p[1]) for p in entity.points]
    else:
        points = [(p[0], p[1]) for p in entity.points()]
    i = 0
    if isinstance(entity.vertices, list):
        vertices = entity.vertices
    else:
        vertices = entity.vertices()

    for v in vertices:
        if hasattr(v, 'bulge'):
            b = v.bulge
        else:
            b = v.get_dxf_attrib('bulge', 0.0)
        p1 = points[i]
        try:
            p2 = points[i+1]
        except Exception:
            if not entity.is_closed:
                break
            p2 = points[0]
        if b != 0.0:
            dx, dy = p2[0] - p1[0], p2[1] - p1[1]
            c = np.sqrt(dx**2 + dy**2)
            s = b * c/2
            r = ((c/2)**2 + s**2)/2/s
            g = np.arctan2(dx, dy)
            a = 2*np.arctan(b)
            pc = (p1[0] + r*np.sin(g - np.pi/2 + a),
                  p1[1] + r*np.cos(g - np.pi/2 + a))
            if b < 0:
                sa = np.arctan2(p2[1]-pc[1], p2[0]-pc[0])
                ea = np.arctan2(p1[1]-pc[1], p1[0]-pc[0])
            else:
                sa = np.arctan2(p1[1]-pc[1], p1[0]-pc[0])
                ea = np.arctan2(p2[1]-pc[1], p2[0]-pc[0])
            logger.debug("Poly p1 %s p2 %s r %s", p1, p2, r)
            yield Arc(Element(center=(pc[0], pc[1]),
                              radius=np.abs(r),
                              start_angle=sa/rf,
                              end_angle=ea/rf),
                      lf, rf,
                      xoff=xoff, yoff=yoff,
                      rotation=rotation)
        else:
            yield Line(Element(start=p1, end=p2), lf,
                       xoff=xoff, yoff=yoff,
                       rotation=rotation)
        i += 1


def lw_polyline(entity, lf, xoff=0.0, yoff=0.0, rotation=0.0):
    """returns a collection of bulged vertices
    http://www.afralisp.net/archive/lisp/Bulges1.htm
    """
    if isinstance(entity.points, list):
        points = [(lf*p[0], lf*p[1]) for p in entity.points]
    else:
        points = [(lf*p[0], lf*p[1]) for p in entity.points()]

    if points:
        p1 = points[0]
        for p2 in points[1:]:
            yield Line(Element(start=p1, end=p2), lf,
                       xoff=xoff, yoff=yoff,
                       rotation=rotation)
            p1 = p2
    if entity.is_closed:
        yield Line(Element(start=p1, end=points[0]), lf,
                   xoff=xoff, yoff=yoff,
                   rotation=rotation)


def ellipse(entity, lf, xoff=0.0, yoff=0.0, rotation=0.0):
    w = np.linalg.norm(entity.major_axis) * 2
    h = entity.ratio * w
    theta = np.arctan2(entity.major_axis[1], entity.major_axis[0])
    start_angle = entity.start_param
    end_angle = entity.end_param
    if end_angle < start_angle:
        end_angle += 2*np.pi
    alfa = np.linspace(start_angle, end_angle, 20)
    x = 0.5 * w * np.cos(alfa)
    y = 0.5 * h * np.sin(alfa)
    R = np.array([
        [np.cos(theta), -np.sin(theta)],
        [np.sin(theta),  np.cos(theta)]
    ])
    x, y = np.dot(R, [x, y])
    x += entity.center[0]
    y += entity.center[1]
    points = np.array((x, y)).T
    p1 = points[0]
    for p2 in points[1:]:
        yield Line(Element(start=p1, end=p2), lf,
                   xoff=xoff, yoff=yoff,
                   rotation=rotation)
        p1 = p2


def spline(entity, lf, min_dist=0.001, xoff=0.0, yoff=0.0, rotation=0.0):
    if False:
        yield Line(Element(start=entity.control_points[0],
                           end=entity.control_points[-1]), lf,
                   xoff=xoff, yoff=yoff,
                   rotation=rotation)
        return

    if False:
        p_prev = None
        for p in entity.control_points:
            if p_prev:
                yield Line(Element(start=p_prev, end=p), lf,
                           xoff=xoff, yoff=yoff,
                           rotation=rotation)
            p_prev = p
        return

    points_between = entity.control_points[1:-1]
    p1 = entity.control_points[0]
    pe = entity.control_points[-1]
    for p2 in points_between:
        dist_12 = distance(p1, p2)
        dist_2e = distance(p2, pe)
        if dist_2e < min_dist:
            logger.debug("SPLINE: ignor small end-distance %s", dist_2e)
            yield Line(Element(start=p1, end=pe), lf,
                       xoff=xoff, yoff=yoff,
                       rotation=rotation)
            return

        if dist_12 > min_dist:
            yield Line(Element(start=p1, end=p2), lf,
                       xoff=xoff, yoff=yoff,
                       rotation=rotation)
            p1 = p2
        else:
            logger.debug("SPLINE: ignor small distance %s", dist_12)

    yield Line(Element(start=p1, end=pe), lf,
               xoff=xoff, yoff=yoff,
               rotation=rotation)


def face3d(entity, lf):
    logger.info("FACE3D: Points=%s", entity.points)
    for i in range(len(entity.points)-1):
        if not entity.is_edge_invisible(i):
            ip = i+1 if i < 4 else 0
            yield Line(Element(start=(entity.points[i][1],
                                      entity.points[i][2]),
                               end=(entity.points[ip][1],
                                    entity.points[ip][2])))


def insert_block(dwg, insert_entity, lf, rf, block, min_dist=0.001):
    logger.debug('Insert %s entities from block %s',
                 len(block),
                 insert_entity.name)
    logger.debug('Insert = %s', insert_entity.insert)
    logger.debug('Rotation = %s', insert_entity.rotation)
    logger.debug('Scale = %s', insert_entity.scale)
    logger.debug('Rows = %s', insert_entity.row_count)
    logger.debug('Cols = %s', insert_entity.col_count)
    logger.debug('Row spacing = %s', insert_entity.row_spacing)
    logger.debug('Col spacing = %s', insert_entity.col_spacing)

    if insert_entity.insert != (0.0, 0.0, 0.0):
        logger.debug('Different Location in Insert')

    xoff = insert_entity.insert[0]
    yoff = insert_entity.insert[1]

    scale = (round(insert_entity.scale[0], 8),
             round(insert_entity.scale[1], 8),
             round(insert_entity.scale[2], 8))
    if not (scale == (1.0, 1.0, 1.0) or
            scale == (1.0, 1.0, 0.0)):
        logger.error('Block scaling in Insert not supported')
        logger.error('  scale = {}'.format(scale))
        return

    if(insert_entity.row_count > 1 or
       insert_entity.col_count > 1 or
       insert_entity.row_spacing > 0 or
       insert_entity.col_spacing > 0):
        logger.error('Multi Block references in Insert not supported')
        return

    for e in block:
        if e.dxftype == 'ARC':
            logger.debug("Block Arc")
            yield Arc(e, lf, rf,
                      xoff=xoff, yoff=yoff,
                      rotation=insert_entity.rotation)
        elif e.dxftype == 'CIRCLE':
            logger.debug("Block Circle %s, Radius %f", e.center[:2], e.radius)
            yield Circle(e, lf,
                         xoff=xoff, yoff=yoff,
                         rotation=insert_entity.rotation)
        elif e.dxftype == 'LINE':
            logger.debug("Block Line")
            yield Line(e, lf,
                       xoff=xoff, yoff=yoff,
                       rotation=insert_entity.rotation)
        elif e.dxftype == 'POLYLINE':
            logger.debug("Block Polyline")
            for p in polylines(e, lf, rf,
                               xoff=xoff, yoff=yoff,
                               rotation=insert_entity.rotation):
                yield p
        elif e.dxftype == 'LWPOLYLINE':
            logger.debug("Block lwPolyline")
            for p in lw_polyline(e, lf,
                                 xoff=xoff, yoff=yoff,
                                 rotation=insert_entity.rotation):
                yield p
        elif e.dxftype == 'SPLINE':
            logger.debug("Block spline")
            for l in spline(e, lf,
                            min_dist=min_dist,
                            xoff=xoff, yoff=yoff,
                            rotation=insert_entity.rotation):
                yield l
        elif e.dxftype == 'INSERT':
            logger.debug("Nested Insert of Block %s", e.name)
            block = dwg.blocks[e.name]
            for l in insert_block(dwg, e, lf, rf, block, min_dist=min_dist):
                yield l

        else:
            logger.warn("unknown type %s in block %s",
                        e.dxftype, insert_entity.name)


def dxfshapes0(dxffile, mindist=0.01, layers=[]):
    """returns a collection of dxf entities (ezdxf)"""
    import ezdxf
    dwg = ezdxf.readfile(dxffile)
    id = 0
    # $ACADVER: AC1006 = R10, AC1009 = R11 and R12, AC1012 = R13,
    #   AC1014 = R14 AC1015 = Release 2000/0i/2
    # check units:
    # dwg.header['$ANGDIR'] 1 = Clockwise angles, 0 = Counterclockwise
    # dwg.header['$AUNITS'] 0 Decimal Degrees, 1 Deg/Min/Sec, 2 Grads, 3 Radians
    # dwg.header['$INSUNIT'] 1 = Inches; 2 = Feet; 3 = Miles;
    #   4 = Millimeters; 5 = Centimeters; 6 = Meters
    # dwg.header['$LUNITS']
    for e in dwg.modelspace():
        if e.dxftype() == 'ARC':
            yield Arc(e.dxf)
        elif e.dxftype() == 'CIRCLE':
            logger.debug("Circle %s, Radius %f", e.center[:2], e.radius)
            yield Circle(e.dxf)
        elif e.dxftype() == 'LINE':
            yield Line(e.dxf)
        elif e.dxftype() == 'POLYLINE':
            for p in polylines(e):
                yield p
        elif e.dxftype() == 'SPLINE':
            for l in spline(e, 1.0, in_dist=mindist):
                yield l
        elif e.dxftype() == 'POINT':
            logger.debug("Id %d4: type %s ignored", id, e.dxftype)
        else:
            logger.warning("Id %d4: unknown type %s", id, e.dxftype)
        id += 1


def dxfshapes(dxffile, mindist=0.01, layers=[]):
    """returns a collection of dxf entities (dxfgrabber)"""
    import dxfgrabber
    dwg = dxfgrabber.readfile(dxffile)
    # print("Layers = {}".format(dwg.layers.names()))
    id = 0
    # $ACADVER: AC1006 = R10, AC1009 = R11 and R12, AC1012 = R13,
    #   AC1014 = R14 AC1015 = Release 2000/0i/2
    # check units:
    # dwg.header['$ANGDIR'] 1 = Clockwise angles, 0 = Counterclockwise
    # dwg.header['$AUNITS'] Decimal Degrees, Deg/Min/Sec, Grads, Radians
    # dwg.header['$INSUNIT'] 1 = Inches; 2 = Feet; 3 = Miles;
    #   4 = Millimeters; 5 = Centimeters; 6 = Meters
    # dwg.header['$LUNITS']
    lf = 1
    if dwg.header.get('$LUNITS', 0) == 1:
        # conv = [1, 2.54e-2, 10.12, 633.0, 1e-3, 1e-2, 1]
        lf = 2.54e3

    rf = np.pi/180
    if dwg.header.get('$AUNITS', 0) == 4:
        rf = 1

    for e in dwg.modelspace():
        if not layers or e.layer in layers:
            if e.dxftype == 'ARC':
                yield Arc(e, lf, rf)
            elif e.dxftype == 'CIRCLE':
                logger.debug("Circle %s, Radius %f", e.center[:2], e.radius)
                yield Circle(e, lf)
            elif e.dxftype == 'LINE':
                yield Line(e, lf)
            elif e.dxftype == 'POLYLINE':
                for p in polylines(e, lf, rf):
                    yield p
            elif e.dxftype == 'LWPOLYLINE':
                for p in lw_polyline(e, lf):
                    yield p
            elif e.dxftype == 'SPLINE':
                for l in spline(e, lf, min_dist=mindist):
                    yield l
            elif e.dxftype == 'INSERT':
                logger.debug("Insert of Block %s", e.name)
                block = dwg.blocks[e.name]
                for l in insert_block(dwg, e, lf, rf, block, min_dist=mindist):
                    yield l
            elif e.dxftype == 'ELLIPSE':
                for l in ellipse(e, lf):
                    yield l
                #w = np.linalg.norm(e.major_axis) * 2
                #h = e.ratio * w
                #rtheta = np.arctan2(e.major_axis[1], e.major_axis[0])
                #angle = rtheta*180/np.pi
                #start_angle = e.start_param*180/np.pi + angle
                #end_angle = e.end_param*180/np.pi + angle
                #if end_angle < start_angle:
                #    end_angle += 360
                #arc = Arc(Element(center=e.center,
                #                  radius=w/2,
                #                  start_angle=start_angle,
                #                  end_angle=end_angle,
                #                  width=w,
                #                  height=h,
                #                  rtheta=rtheta,
                #                  start_param=e.start_param,
                #                  end_param=e.end_param))
                #yield arc

            elif e.dxftype == 'POINT':
                logger.debug("Id %d4: type %s ignored", id, e.dxftype)
            elif e.dxftype == '3DFACE':
                logger.warning(
                    "Id %d4: type %s not implemented", id, e.dxftype)
                # for l in face3d(e, lf):
                #     yield l
            else:
                logger.warning("Id %d4: unknown type %s", id, e.dxftype)
            id += 1



# -*- coding: utf-8 -*-
'''Calculate leakage inductances'''
import numpy as np
import logging

logger = logging.getLogger(__name__)


def end_wdg_leak_ind_round_wires(l_ew, p, q, m, y, taup, w1, layers, num_par_wdgs, slot_height):
    '''returns end winding leakage inductance per phase per end winding side (DE or NDE)'''
    mue0 = 4*np.pi*1e-7
    if layers == 2:
        lambda_ew = 0.34*q*(1 - 2*y*taup/(np.pi*l_ew*m*q))
    if layers == 1:
        lambda_ew = q*(0.67 - 0.43*(taup + slot_height*np.pi/(2*p))/l_ew)
    L_ew = mue0*2/(p*q)*(w1/num_par_wdgs)**2*l_ew*lambda_ew
    return L_ew


def end_wdg_leak_ind_hairpin_wires(): #TODO
    L_ew = 0
    return L_ew


def harm_leak_ind(E_fft, order_fft, freq, Ia): # needs to be validated
    '''returns harmonic leakage inductance per phase'''
    L_harm = []
    for ii in range(1, len(order_fft)): # skip 1st order
        L_harm.append(E_fft[ii] / Ia / (2*np.pi*freq*order_fft[ii]))

    return sum(L_harm)




""" Axial Flux PM Machine

"""
import logging
import numpy as np
from pathlib import Path
import shutil
from .. import poc
from .. import parstudy
from .. import model
from .. import utils
from .. import windings
from .. import femag
from .. import ecloss
from .. import nc
from .. import bch

from scipy.interpolate import make_interp_spline, RegularGridInterpolator, RectBivariateSpline
from scipy.integrate import quad
import copy
from matplotlib.colors import to_rgb
from multiprocessing import Pool

logger = logging.getLogger(__name__)

AFM_TYPES = (
    "S1R1",      # 1 stator, 1 rotor
    "S1R2",      # 1 stator, 2 rotor, 1 half simulated
    "S1R2_all",  # 1 stator, 2 rotor, all simulated
    "S2R1",      # 2 stator, 1 rotor, 1 half simulated
    "S2R1_all"   # 2 stator, 1 rotor, all simulated
)

def stator_template_name(machine):
    for k in machine['stator']:
        if type(machine['stator'][k]) == dict:
            return k
    return ''

def magnet_template_name(machine):
    for k in machine['magnet']:
        if type(machine['magnet'][k]) == dict:
            return k
    return ''

def jsonify(v):
    vtype = type(v)
    if vtype == np.float32:
        return float(v)
    if vtype == np.ndarray:
        return v.tolist()
    if vtype == dict:
        for k in v:
            v[k] = jsonify(v[k])
    return v


def get_magdata(task):
    """result func used in parstudy
    extracts the magnet data from the NC/ISA7 for
    subsequent magnet loss calc
    """
    basedir = Path(task.directory)
    bchfile_list = sorted(basedir.glob(
        '*_[0-9][0-9][0-9].B*CH'))
    if bchfile_list:
        result = bch.Reader()
        with open(bchfile_list[-1]) as f:
            logger.debug("Reading %s",
                         bchfile_list[-1])
            result.read(f)
    fnc = list(basedir.glob('*.nc'))[0]
    ibeta = len(result.linearForce)-1
    ncmod = nc.read(fnc)
    result.losses[0]['magnet_data'] = [jsonify(pm)
                                       for pm in ncmod.get_magnet_data(
                                               ibeta=ibeta)]
    return result


def num_agnodes(Q, p, pw, ag):
    """return total number of nodes in airgap per pole
     Args:
      Q: (int) number of slots
      p: (int) number of pole pairs
      pw: (float) pole width
      ag: (float) airgap height
    """
    num_nodes = np.arange(6, 120, 6)
    i = np.argmin(np.abs(pw/num_nodes - ag/2))
    nag = num_nodes[i-1]
    if p*nag % Q:
        lcm = np.lcm(Q, 2*p)//p
        nmin, nmax = 1, num_nodes[-1]//lcm
        num_nodes = np.array(
            [i*lcm for i in range(nmin, nmax) if i*lcm % 6 == 0])
        i = np.argmin(np.abs(pw/num_nodes - ag/2))
        if i > 0:
            nag = num_nodes[i-1]
        else:
            nag = num_nodes[i]
        # TODO nodedist 0.5, 2, 4, 6
    return nag


def _integrate(radius, pos, val):
    interp = RegularGridInterpolator((radius, pos), val)
    def func(x, y):
        return interp((x, y))
    return [quad(func, radius[0], radius[-1],
                  args=(p,), limit=200)[0]
            for p in pos]


def _integrate1d(radius, val):
    interp = make_interp_spline(radius, val, k=1)
    def func(x):
        return interp((x))
    return quad(func, radius[0], radius[-1], limit=100)[0]


def ld_interpol(i1, beta, v):
    '''interpolate Ld at beta angle 0°, -180°'''
    # ld
    cur = copy.deepcopy(i1)
    betad = copy.deepcopy(beta)
    if np.amin(beta) < -90 and \
        np.amax(beta) > -90:
        # motor and generator
        v[0] = v[1]
        v[-1] = v[-2]

        dbeta = np.abs(beta[0][0] - beta[1][0])
        bp = [[beta[0][0]-dbeta for i in range(len(np.unique(i1)))]] + beta[1:-1] + \
            [[dbeta for i in range(len(np.unique(i1)))]]
    else:
        v[-1] = v[-2]
        dbeta = np.abs(beta[0][0] - beta[1][0])
        bp = beta[0:-1] + \
            [[dbeta for i in range(len(np.unique(i1)))]]

    return RectBivariateSpline(np.unique(bp), np.unique(cur),
                               np.array(v)).ev(*[betad, i1]).tolist()


def lq_interpol(i1, beta, v):
    '''interpolate Lq at beta -90°'''
    if -90 not in np.unique(beta):
        return v
    # lq
    betad = copy.deepcopy(beta)
    if np.amin(beta) < -90 and \
        np.amax(beta) > -90:
        # motor and generator
        inx = np.argwhere(np.array(beta) == -90).squeeze()
        v.pop(inx[0, 0])
        bp = beta[0:inx[0, 0]] + beta[inx[0, 0]+1:]
        cp = i1[0:inx[0, 0]] + i1[inx[0, 0]+1:]
    else:
        v[0] = v[1]
        dbeta = np.abs(beta[0][0] - beta[1][0])
        bp = [[-90-dbeta for i in i1[0]]] + beta[1::]
        cp = i1
    cur = copy.deepcopy(cp)
    return RectBivariateSpline(np.unique(bp), np.unique(cur), \
         np.array(v)).ev(*[betad, i1]).tolist()


def parident(workdir, engine, temp, machine,
             magnetizingCurves, magnetMat=[], condMat=[],
             **kwargs):
    """return list of parameters of equivalent circuit for AFPM

    arguments:
    workdir: (str) directory for intermediate files
    engine: (object) calculation driver (multiproc, docker, condor)

    temp: list of magnet temperatures in degree Celsius
    machine: dict() with machine parameters
    magnetizingCurves: list of dict() with BH curves
    magnetMat: list of dict() with magnet material properties
    condMat: list of dict() with conductor material properties

    optional arguments:
    num_slices: number of slices (default 3)
    num_cur_steps: number of current steps (default 5)
    num_beta_steps: number of current steps (default 7 per quadrant)
    speed: rotor speed in 1/s (default 160/p)
    i1_max: maximum current in A rms (default approx 3*i1nom)
    cmd: femag executable
    """
    try:
        if machine['afmtype'] not in AFM_TYPES:
            raise ValueError(f"invalid afm type {machine['afmtype']}")
    except KeyError:
        raise ValueError("missing key afmtype")

    di = machine['inner_diam']
    Q1 = machine['stator']['num_slots']
    slotmodel = stator_template_name(machine)
    hs = machine['stator'][slotmodel]['slot_height']
    wdgk = 'windings' if 'windings' in machine else 'winding'
    N = machine[wdgk]['num_wires']
    Jmax = 15  # max current density in A/mm2

    i1_max = kwargs.get(
        'i1_max',
        round(0.28*np.pi*hs*(di+hs)/Q1/N*Jmax*1e5)*10 * \
        machine[wdgk].get('num_par_wdgs', 1))

    p = machine['poles']
    slotmodel = magnet_template_name(machine)
    if np.isscalar(machine['magnet'][slotmodel]['rel_magn_width']):
        num_slices = kwargs.get('num_slices', 3)
        rmagw = num_slices*[machine['magnet'][slotmodel]['rel_magn_width']]
    elif len(machine['magnet'][slotmodel]['rel_magn_width']) == 1: 
            num_slices = kwargs.get('num_slices', 3)
            rmagw = num_slices*machine['magnet'][slotmodel]['rel_magn_width']
    else:
        rmagw = machine['magnet'][slotmodel]['rel_magn_width']
        num_slices = len(rmagw)

    logger.info("num_slices %d rmagw %s", num_slices, rmagw)
    lfe = get_arm_lengths(machine['outer_diam'],
                          machine['inner_diam'],
                          num_slices)
    pole_width = get_pole_widths(machine['outer_diam'],
                                 machine['inner_diam'],
                                 p,
                                 num_slices)

    speed = kwargs.get('speed', 160/p)
    linspeed = [speed*p*pw for pw in pole_width]

    if "num_agnodes" not in machine:
        for pw in pole_width:
            machine['num_agnodes'] = num_agnodes(Q1, p//2, pw,
                                                 machine['airgap'])
    nlparvardef = {
        "decision_vars": [
            {"values": pole_width,
             "name": "pole_width"},
            {"values": rmagw,
             "name": f"magnet.{slotmodel}.rel_magn_width"},
            {"values": lfe,
             "name": "lfe"},
            {"values": linspeed, "name": "speed"}
        ]
    }
    machine['pole_width'] = np.pi * machine['inner_diam']/machine['poles']
    machine['lfe'] = machine['outer_diam'] - machine['inner_diam']

    beta_min = kwargs.get('beta_min', -90)
    default_steps = round(-beta_min/15)+1
    num_cur_steps = kwargs.get('num_cur_steps', 5)
    num_beta_steps = kwargs.get('num_beta_steps', default_steps)
    parvardef = {
        "decision_vars": [
            #            {"values": sorted(2*temp), "name": "magn_temp"},
            {"steps": num_beta_steps, "bounds": [beta_min, 0],
             "name": "angl_i_up"},
            {"steps": num_cur_steps,
             "bounds": [i1_max/num_cur_steps, i1_max], "name": "current"}
        ]
    }

    ldq = []
    for magtemp in temp:
        nlcalc = dict(
            calculationMode="cogg_calc",
            num_move_steps=60,
            magn_temp=magtemp,
            poc=poc.Poc(999),
            speed=0)
        logging.info("Noload simulation")
        if kwargs.get('use_multiprocessing', True):
            pstudy = parstudy.List(
                    workdir, condMat=condMat, magnets=magnetMat,
                    magnetizingCurves=magnetizingCurves,
                    cmd=kwargs.get('cmd', None))

            nlresults = pstudy(nlparvardef, machine, nlcalc, engine)
            if nlresults['status'].count('C') != len(nlresults['status']):
                raise ValueError(
                    f"Noload simulation failed {nlresults['status']}")
        else:
            nlresults = {"x": [], "f": []}
            i = 0

            for pw, le, sp, rmw in zip(pole_width, lfe, linspeed, rmagw):
                nlmachine = {k: machine[k] for k in machine}
                nlmachine['pole_width'] = pw
                nlmachine['magnet'][slotmodel]['rel_magn_width'] = rmw
                nlmachine['lfe'] = le
                nlcalc.update({"speed": sp})
                nlsubdir = f'{workdir}/{i}'
                nlworkdir = Path(nlsubdir)
                if nlworkdir.exists():
                    shutil.rmtree(nlworkdir)
                nlworkdir.mkdir(exist_ok=True)
                noloadsim = femag.Femag(
                    nlworkdir, condMat=condMat, magnets=magnetMat,
                    magnetizingCurves=magnetizingCurves,
                    cmd=kwargs.get('cmd', None))
                r = noloadsim(nlmachine, nlcalc)
                nlresults['f'].append({k: v for k, v in r.items()})
                i = i + 1
        nlresults.update(process(lfe, pole_width, machine, nlresults['f']))
        weights = nlresults['weights']
        current_angles = nlresults['f'][0]['current_angles']
        results = []
        i = 0
        for l, pw, rmw in zip(lfe, pole_width, rmagw):
            mpart = {k: machine[k] for k in machine if k != slotmodel}
            mpart['pole_width'] = pw
            mpart['lfe'] = l
            mpart['magnet'][slotmodel]['rel_magn_width'] = rmw
            subdir = f"{workdir}/{i}"

            simulation = dict(
                calculationMode="torq_calc",
                wind_temp=20.0,
                magn_temp=magtemp,
                angl_i_up=0.0,
                magn_height=machine['magnet'][slotmodel]['magn_height'],
                yoke_height=machine['magnet'][slotmodel].get(
                    'yoke_height', 0),
                current=1,
                poc=poc.Poc(999,
                            parameters={
                                'phi_voltage_winding': current_angles}),
                num_move_steps=60,
                speed=linspeed[i],
                num_par_wdgs=machine[wdgk].get('num_par_wdgs', 1))

            if kwargs.get('use_multiprocessing', True):
                gpstudy = parstudy.Grid(
                    subdir, condMat=condMat, magnets=magnetMat,
                    magnetizingCurves=magnetizingCurves,
                    cmd=kwargs.get('cmd', None),
                    result_func=get_magdata)
                lresults = gpstudy(parvardef, mpart, simulation, engine)

            else:
                lresults = {"x": [], "f": []}
                domain_beta = np.linspace(beta_min, 0, num_beta_steps).tolist()
                domain_cur = np.linspace(i1_max/num_cur_steps, i1_max, num_cur_steps).tolist()
                dir_index = 0
                for cur in domain_cur:
                    for be in domain_beta:
                        simulation['angl_i_up'] = be
                        simulation['current'] = cur
                        lresults['x'].append([be, cur])
                        subsubdir = subdir + f'/{dir_index}'
                        dir_index = dir_index + 1
                        lworkdir = Path(subsubdir)
                        if lworkdir.exists():
                            shutil.rmtree(lworkdir)
                        lworkdir.mkdir(exist_ok=True)
                        loadsim = femag.Femag(
                            lworkdir, condMat=condMat, magnets=magnetMat,
                            magnetizingCurves=magnetizingCurves,
                            cmd=kwargs.get('cmd', None))
                        r = loadsim(mpart, simulation)
                        lresults['f'].append({k: v for k, v in r.items()})

            f = [{k: bch[k]
                  for k in ('linearForce', 'flux', 'losses', 'lossPar')}
                 for bch in lresults['f']]
            lresults['f'] = f
            results.append(lresults)
            i += 1

        postp = []
        if kwargs.get('use_multiprocessing', True):
            with Pool() as p:
                for r in p.starmap(process,
                                   [(lfe, pole_width, machine, bch)
                                    for bch in zip(*[r['f']
                                                     for r in results])]):
                    torque = np.mean(r.pop('torque'))
                    r['torque'] = torque
                    r.update(_psidq_ldq(r, nlresults))
                    postp.append(r)
        else:
            for r in [process(lfe, pole_width, machine, bch)
                      for bch in zip(*[r['f'] for r in results])]:
                torque = np.mean(r.pop('torque'))
                r['torque'] = torque
                r.update(_psidq_ldq(r, nlresults))
                postp.append(r)

        r1 = postp[0]['r1']
        i1 = [r['i1'] for r in postp][::num_beta_steps]
        beta = [r['beta'] for r in postp][:num_beta_steps]
        # Note: Psi RMS Values aligned with BCH/BATCH
        psid = np.reshape([r['psid'] for r in postp],
                          (-1, num_beta_steps)).T/np.sqrt(2)
        psiq = np.reshape([r['psiq'] for r in postp],
                          (-1, num_beta_steps)).T/np.sqrt(2)

        ld = np.reshape([r['Ld'] for r in postp],
                          (-1, num_beta_steps)).T.tolist()
        lq = np.reshape([r['Lq'] for r in postp],
                          (-1, num_beta_steps)).T.tolist()
        # interpolation ld, lq
        curr, angl = [], []
        for cr in range(len(beta)):
            curr.append(i1)
        for al in beta:
            tmp = []
            for cr in range(len(i1)):
                tmp.append(al)
            angl.append(tmp)
        try:
            xx, yy = copy.deepcopy(curr), copy.deepcopy(angl)
            ld = ld_interpol(xx, yy, ld)
            xx, yy = copy.deepcopy(curr), copy.deepcopy(angl)
            lq = lq_interpol(xx, yy, lq)
        except:
            ld = np.zeros_like(psid).tolist()
            lq = np.zeros_like(psid).tolist()

        torque = np.reshape([r['torque'] for r in postp],
                            (-1, num_beta_steps)).T
        losses = {k: np.flip(np.reshape([r['plfe'][k] for r in postp],
                                        (-1, num_beta_steps)),
                             axis=1).T.tolist()
                  for k in postp[0]['plfe']}
        losses.update({'speed': speed,
                       'hf': postp[0]['lossPar']['hf'],
                       'ef': postp[0]['lossPar']['ef'],
                       'magnet': np.flip(
                           np.reshape([r['plmag'] for r in postp],
                                      (-1, num_beta_steps)),
                           axis=1).T.tolist()})
        ldq.append({'temperature': magtemp,
                    'i1': i1, 'beta': beta,
                    'psid': psid.tolist(), 'psiq': psiq.tolist(),
                    'ld': ld, 'lq': lq,
                    'torque': torque.tolist(),
                    'losses': losses})
        # T = 3/2 p (Psid iq - Psiq id)
        #iq, id = femagtools.machine.utils.iqd(*np.meshgrid(beta, i1))

    return {'m': machine[wdgk]['num_phases'],
            'p': machine['poles']//2, 'weights': weights,
            'rotor_mass': sum(weights[1]), "kfric_b": 1,
            'ls1': 0, 'r1': r1, 'ldq': ldq}


def process(lfe, pole_width, machine, bch):
    """process results: torque, voltage (emf), losses

    Arguments:
    lfe: (float) active machine length
    pole_width: (float) pole width
    machine: (dict) machine
    bch: (list of dict) linearForce, flux, losses

    Returns dict with keys:
    pos: (list of float)
    r1: winding resistance
    torque: (list of float)
    emf: (list of float)
    emf_amp: (float)
    emf_angle: (float)
    freq: (float)
    currents: (list of list of float)
    plfe, plmag, plcu: (list of float)
    """
    model_type = machine['afmtype']
    num_slots = machine['stator']['num_slots']
    mmod = model.MachineModel(machine)
    slots_gen = mmod.stator['num_slots_gen']
    scale_factor = _get_scale_factor(model_type, num_slots, slots_gen)
    endpos = [2*pw*1e3 for pw in pole_width]
    displ = [[d for d in r['linearForce'][-1]['displ']
              if d < e*(1+1/len(r['linearForce'][-1]['displ']))]
             for e, r in zip(endpos, bch)]
    radius = [pw*machine['poles']/2/np.pi for pw in pole_width]
    rotpos = [np.array(d)/r/1000 for r, d in zip(radius, displ)]
    n = len(rotpos[0])
    currents = [bch[0]['flux'][k][0]['current_k'][:n]
                for k in bch[0]['flux']]

    if len(pole_width) > 1:
        # check homogenity:
        if np.diff([len(d) for d in displ]).any():
            raise ValueError(
                f"inhomogenous number of steps: {[len(d) for d in displ]}")
        lfx = [r['linearForce'][-1]['force_x'] for r in bch]
        torque = _integrate(radius, rotpos[0], np.array(
            [r*scale_factor*np.array(fx[:-1])/l
             for l, r, fx in zip(lfe, radius, lfx)]))

        voltage = {k: [scale_factor * np.array(ux[:-1])/l
                       for l, ux in zip(lfe, [r['flux'][k][-1]['voltage_dpsi']
                                              for r in bch])]
                   for k in bch[0]['flux']}
        emf = [_integrate(radius, rotpos[0], np.array(voltage[k]))
               for k in voltage]

        fluxxy = {k: [scale_factor * np.array(flx[:-1])/l
                      for l, flx in zip(lfe, [r['flux'][k][-1]['flux_k']
                                              for r in bch])]
                  for k in bch[0]['flux']}
        flux = [_integrate(radius, rotpos[0], np.array(fluxxy[k]))
                for k in fluxxy]
    else:
        r = radius[0]
        torque = [r*scale_factor*fx
                  for fx in bch[0]['linearForce'][-1]['force_x'][:-1]]
        voltage = {k: [scale_factor * ux
                       for ux in bch[0]['flux'][k][-1]['voltage_dpsi'][:-1]]
                   for k in bch[0]['flux']}
        emf = [voltage[k][:n] for k in voltage]
        fluxxy = {k: [scale_factor * np.array(flx)
                      for flx in bch[0]['flux'][k][-1]['flux_k']]
                  for k in bch[0]['flux']}
        flux = [fluxxy[k][:n] for k in fluxxy]

    pos = (rotpos[0]/np.pi*180)
    emffft = utils.fft(pos, emf[0])

    scale_factor = 1 if model_type == 'S1R1' else 2
    styoke = {}
    stteeth = {'hyst': 0, 'eddy': 0}
    rotor = {}
    for k in ('hyst', 'eddy'):
        if len(pole_width) > 1:
            styoke[k] = _integrate1d(radius, scale_factor*np.array(
                [sum(b['losses'][-1]['stator']['stfe'][k])/l
                 for l, b in zip(lfe, bch)]))
            rotor[k] = _integrate1d(radius, scale_factor*np.array(
                [sum(b['losses'][-1]['rotor']['----'][k])/l
                 for l, b in zip(lfe, bch)]))
        else:
            styoke[k] = scale_factor*sum(
                bch[0]['losses'][-1]['stator']['stfe'][k])
            rotor[k] = scale_factor*sum(
                bch[0]['losses'][-1]['rotor']['----'][k])

    if 'magnet_data' in bch[0]['losses'][0]:
        k = 'magnetH'
        nsegx = machine['magnet'].get('num_segments', 1)
        if type(nsegx) == int:
            nsegx = [nsegx]*len(bch)
        elif len(nsegx) == 1 and len(nsegx) < len(bch): 
            nsegx = nsegx*len(bch)
        else: 
            pass 
        for nx, b in zip(nsegx, bch):
            pm = b['losses'][0]['magnet_data']
            magloss = ecloss.MagnLoss(magnet_data=[pm])
            ml = magloss.calc_losses_ialh2(nsegx=nx)
            b["losses"][-1].update({k: ml[0]})
    else:
        k = 'magnetJ'

    if len(pole_width) > 1:
        maglosses = _integrate1d(radius, scale_factor*np.array(
            [b['losses'][-1][k]/lz
             for lz, b in zip(lfe, bch)]))
    else:
        maglosses = scale_factor*bch[0]['losses'][-1][k]

    freq = bch[0]['losses'][-1]['stator']['stfe']['freq'][0]

    wdg = windings.Winding(
        dict(Q=mmod.stator['num_slots'],
             p=mmod.poles//2,
             m=mmod.winding['num_phases'],
             l=mmod.winding['num_layers']))
    slotmodel = stator_template_name(machine)
    cufill = mmod.winding.get('cufilfact', 0.4)
    aw = (mmod.stator[slotmodel]['slot_width']*
              mmod.stator[slotmodel]['slot_height']*
              cufill/mmod.winding['num_wires']/mmod.winding['num_layers'])
    r1 = wdg_resistance(wdg, mmod.winding['num_wires'],
                        mmod.winding['num_par_wdgs'],
                        aw,
                        mmod.outer_diam, mmod.inner_diam)
    i1 = np.mean([np.max(c) for c in currents])/np.sqrt(2)
    plcu = mmod.winding['num_phases']*i1**2*r1
    weights = np.array([[0, 0, 0], [0, 0, 0]])
    try:
        for b in bch:
            weights = weights + b['weights']
        weights *= scale_factor
    except KeyError as exc:
        #logger.warning("missing key %s", exc)
        pass
    return {
        'weights': weights.tolist(),
        'pos': pos.tolist(), 'r1': r1,
        'torque': torque,
        'flux_k': flux,
        'emf': emf,
        'emf_amp': emffft['a'], 'emf_angle': emffft['alfa0'],
        'freq': freq,
        'currents': currents,
        'beta': bch[0]['losses'][0]['beta'],
        'plfe': {'styoke_hyst': styoke['hyst'], 'styoke_eddy': styoke['eddy'],
                 'stteeth_hyst': stteeth['hyst'], 'stteeth_eddy': stteeth['eddy'],
                 'rotor_hyst': rotor['hyst'], 'rotor_eddy': rotor['eddy']},
        'plmag': maglosses, 'lossPar': {k: bch[0]['lossPar'][k]
                                        for k in bch[0]['lossPar']},
        'plcu': plcu}


def _psidq_ldq(results, nlresults):
    """ calculate currents, winding fluxes psi and inductances L
    Arguments:
    results: (dict) emf, currents, freq of load simulation
    nlresults: (dict) emf of no load simulation

    returns dict of currents, psi, L
    psid, psiq, psim, i1, id, iq, Ld, Lq
    """
    gamma = -(results['emf_angle'] - nlresults['emf_angle'])
    w1 = 2*np.pi*results['freq']
    psid = np.cos(gamma)*results['emf_amp']/w1
    psiq = -np.sin(gamma)*results['emf_amp']/w1
    psim = nlresults['emf_amp']/w1
    i1 = np.mean([np.max(c)
                  for c in results['currents']])/np.sqrt(2)
    beta = results['beta']/180*np.pi
    id = np.sqrt(2)*i1*np.sin(beta)
    iq = np.sqrt(2)*i1*np.cos(beta)
    Ld = Lq = float('nan')
    if np.abs(id) > 0:
        Ld = (psid - psim)/id
    if np.abs(iq) > 0:
        Lq = psiq/iq
    return {
        'psid': psid,
        'psiq': psiq,
        'psim': psim,
        'i1': i1,
        'beta': results['beta'],
        'id': id,
        'iq': iq,
        'Ld': Ld,
        'Lq': Lq}

def _get_scale_factor(model_type, num_slots, slots_gen):
    """Determines the scale factor
    Arguments:
    model_type : (str) type of model
    num_slots: (int) total number of stator slots
    slots_gen: (int) number of slots in model

    Returns scale factor based on number of poles, number of poles simulated
        and the model type
    """
    segments = num_slots / slots_gen
    if model_type in ("S2R1", "S1R2"):
        return 2 * segments
    return segments


def get_arm_lengths(outer_diam, inner_diam, num_slices):
    d = outer_diam - inner_diam
    if num_slices > 2:
        return [d/(4*(num_slices-1))] + [
            d/(2*(num_slices-1))
            for i in range(1,num_slices-1)] + [d/(4*(num_slices-1))]
    return [d/2]


def get_pole_widths(outer_diam, inner_diam, poles, num_slices):
    d = outer_diam - inner_diam
    if num_slices > 2:
        return [np.pi * inner_diam/poles] + [
            np.pi * (inner_diam + d*i/(num_slices - 1))/poles
            for i in range(1, num_slices-1)] + [np.pi * outer_diam/poles]
    return [np.pi * (outer_diam+inner_diam)/2/poles]


def wdg_resistance(wdg, n, g, aw, outer_diam, inner_diam,
                   sigma=56e6):
    """return winding resistance per phase in Ohm
    Arguments:
    wdg: (Winding) winding
    n: (int) number of wires per coil side
    g: (int) number of parallel coil groups
    aw: wire cross section area m2
    outer_diam, inner_diam: diameters m
    sigma: (float) conductivity of wire material 1/Ohm m
    """
    # mean length of one turn
    lt = 2.4*((outer_diam-inner_diam)+np.pi/wdg.Q*(outer_diam+inner_diam) + 16e-3)
    return wdg.turns_per_phase(n, g)*lt/sigma/aw/g


def _get_copper_losses(scale_factor, bch):
    """return copper losses from bch files"""
    try:
        wdgk = 'winding'
        cu_losses = sum([b['losses'][0][wdgk] for b in bch])
        return scale_factor*cu_losses
    except KeyError:
        return 0  # noload calc has no winding losses


def _set_plot_attributes(ax):
    ax.set_aspect('equal')
    for loc, spine in ax.spines.items():
        spine.set_color('none')  # don't draw spine
    ax.yaxis.set_ticks([])
    ax.xaxis.set_ticks([])


def _get_colors(colors, delta):
    if delta == 0.0:
        return colors
    new_colors = []
    for col in colors:
        rgb = to_rgb(col)
        r, g, b = rgb
        col = (max(0.0, min(r+delta, 1.0)),
               max(0.0, min(g+delta, 1.0)),
               max(0.0, min(b+delta, 1.0)))
        new_colors.append(col)
    return new_colors


def _draw_vertical_magnets(ax,
                           poles,
                           xr, yr,
                           Rr,
                           xoff, yoff,
                           delta=0.0):
    color = ['green', 'red']
    color = _get_colors(color, delta)
    for i in range(poles):
        ax.fill(xr+xoff, yr+yoff,
                facecolor=color[i%2], edgecolor=color[i%2])
        xr, yr = np.dot(Rr, [xr, yr])
    return


def _draw_vertical_slots(ax,
                         Q,
                         r,
                         alpha,
                         xoff, yoff,
                         delta=0.0):
    color = ['skyblue', 'blue']
    color = _get_colors(color, delta)
    taus = 2*np.pi/Q
    for n in range(Q):
        beta0 = np.linspace(n*taus, n*taus + taus/2-alpha[0], 5)
        beta1 = np.linspace(n*taus, n*taus + taus/2-alpha[1], 5)
        xr = np.concatenate((
            r[0]*np.cos(beta0), r[1]*np.cos(beta1[::-1])))+xoff
        yr = np.concatenate((
            r[0]*np.sin(beta0), r[1]*np.sin(beta1[::-1])))+yoff
        ax.fill(xr, yr, color=color[0])
        beta0 = np.linspace(n*taus + taus/2+alpha[0], (n+1)*taus, 5)
        beta1 = np.linspace(n*taus + taus/2+alpha[1], (n+1)*taus, 5)
        xr = np.concatenate((
            r[0]*np.cos(beta0), r[1]*np.cos(beta1[::-1])))+xoff
        yr = np.concatenate((
            r[0]*np.sin(beta0), r[1]*np.sin(beta1[::-1])))+yoff
        ax.fill(xr, yr, color=color[0])


def vertical_plot(machine, ax):
    """plots afpm stator and rotor (vertical section)
    Args:
      dy1, dy1: float outer, inner diameter
      rel_magn_width: float rel magnet width 0..1
      Q: number of stator slots
      poles: number of poles
      slot_width: width of stator slot
    """
    logger.debug("begin of vertical_plot()")

    model_type = machine['afmtype'][0:4]
    dy1 = machine['outer_diam']*1e3
    dy2 = machine['inner_diam']*1e3
    try:
        rel_magn_width = max(machine['magnet']['afm_rotor']['rel_magn_width'])
    except TypeError:
        rel_magn_width = machine['magnet']['afm_rotor']['rel_magn_width']
    Q = machine['stator']['num_slots']
    slot_width = machine['stator']['afm_stator']['slot_width']*1e3
    poles = machine['poles']

    # prepare Magnets
    theta = np.linspace(np.pi/poles*(1-rel_magn_width),
                        np.pi/poles*(1+rel_magn_width),
                        10)
    xr = np.concatenate((dy1/2 * np.cos(theta), dy2/2 * np.cos(theta[::-1])))
    yr = np.concatenate((dy1/2 * np.sin(theta), dy2/2 * np.sin(theta[::-1])))
    rtheta = 2*np.pi/poles
    Rr = np.array([
        [np.cos(rtheta), -np.sin(rtheta)],
        [np.sin(rtheta),  np.cos(rtheta)]
    ])

    # prepare Slots
    taus = 2*np.pi/Q
    r = np.array([dy2/2, dy1/2])
    alpha = np.arctan2(slot_width/2, r)

    yoff = 0.0
    xoff = 0.0
    y_shift = -2
    x_shift = dy1-dy2
    ma_delta = 0.0  # color
    sl_delta = 0.0  # color

    # Draw
    if model_type in ("S1R2"):  # 2 rotor
        _draw_vertical_magnets(ax, poles, xr, yr, Rr, xoff, yoff, delta=-0.1)
        yoff += y_shift
        xoff += x_shift

    if model_type in ("S2R1"):  # 2 stator
        sl_delta = -0.1

    _draw_vertical_slots(ax, Q, r, alpha, xoff, yoff, delta=sl_delta)
    yoff += y_shift
    xoff += x_shift

    if model_type in ("S1R2"):  # 2 rotor
        ma_delta = 0.1

    _draw_vertical_magnets(ax, poles, xr, yr, Rr, xoff, yoff, delta=ma_delta)
    yoff += y_shift
    xoff += x_shift

    if model_type in ("S2R1"):  # 2 stator
        sl_delta = 0.0
        _draw_vertical_slots(ax, Q, r, alpha, xoff, yoff, delta=sl_delta)

    _set_plot_attributes(ax)
    logger.debug("end of vertical_plot()")


IRON_NO = 0
IRON_UP = 1
IRON_DOWN = 2

def _draw_horizontal_magnets(ax,
                             poles,
                             magn_height,
                             magn_width,
                             yoke_height,
                             Q,
                             g,
                             taus,
                             dy2,
                             yoff=0.0,
                             iron=IRON_NO
                             ):
    color = ['green', 'red']
    xy = (0, Q//g*taus, Q//g*taus, 0)

    if iron == IRON_UP:
        yy = (yoff-yoke_height,
              yoff-yoke_height,
              yoff,
              yoff)
        yoff -= yoke_height
        ax.fill(xy, yy, color='skyblue')

    taum = dy2*np.pi/poles
    ym = np.array([yoff-magn_height,
                   yoff-magn_height,
                   yoff,
                   yoff])
    yoff -= magn_height

    for n in range(poles//g):
        xl = taum*n + taum*(1 - magn_width)
        xr = taum*(n + 1) - taum*(1 - magn_width)
        xm = (xl, xr, xr, xl)
        ax.fill(xm, ym, color=color[n%2])

    if iron == IRON_DOWN:
        yy = (yoff-yoke_height,
              yoff-yoke_height,
              yoff,
              yoff)
        yoff -= yoke_height
        ax.fill(xy, yy, color='skyblue')
    return yoff


TOOTH_UP = 0
TOOTH_DOWN = 1
TOOTH_ONLY = 2


def _draw_horizontal_slots(ax,
                           slot_height, slot_width, yoke_height,
                           Q, g, taus,
                           yoff=0.0,
                           tooth=TOOTH_DOWN):
    if not tooth == TOOTH_ONLY:
        xx = (0, Q//g*taus, Q//g*taus, 0)
        if tooth == TOOTH_DOWN:
            yy = (yoff-yoke_height,
                  yoff-yoke_height,
                  yoff,
                  yoff)
        else:
            yy = (yoff-slot_height,
                  yoff-slot_height,
                  yoff-slot_height-yoke_height,
                  yoff-slot_height-yoke_height)
        ax.fill(xx, yy, color='skyblue')

    yt = (yoff-slot_height-yoke_height,
          yoff-slot_height-yoke_height,
          yoff, yoff)
    for n in range(Q//g):
        xt = np.array((n*taus, n*taus+(taus-slot_width)/2,
                      n*taus+(taus-slot_width)/2, n*taus))
        ax.fill(xt, yt, color='skyblue')
        xt += slot_width + (taus-slot_width)/2
        ax.fill(xt, yt, color='skyblue')
    return yoff - slot_height - yoke_height


def horizontal_plot(machine, ax):
    logger.debug("begin of horizontal_plot()")

    model_type = machine['afmtype'][0:4]
    dy1 = machine['outer_diam']*1e3
    dy2 = machine['inner_diam']*1e3
    try:
        rel_magn_width = max(machine['magnet']['afm_rotor']['rel_magn_width'])
    except TypeError:
        rel_magn_width = machine['magnet']['afm_rotor']['rel_magn_width']
    magn_height = machine['magnet']['afm_rotor']['magn_height']*1e3
    magn_yoke_height = machine['magnet']['afm_rotor']['yoke_height']*1e3

    Q = machine['stator']['num_slots']
    slot_width = machine['stator']['afm_stator']['slot_width']*1e3
    poles = machine['poles']
    m = 3
    slot_height = machine['stator']['afm_stator']['slot_height']*1e3
    if model_type in ('S2R1', 'S2R1_all'):
        slot_height /= 2
    yoke_height = machine['stator']['afm_stator']['yoke_height']*1e3
    ag = machine['airgap']*1e3

    g = np.gcd(Q, m*poles)//m
    taus = dy2*np.pi/Q

    yoff = 0.0
    if model_type in ('S1R2', 'S1R2_all'):  # 2 rotor
        yoff = _draw_horizontal_magnets(ax, poles,
                                        magn_height, rel_magn_width,
                                        magn_yoke_height,
                                        Q, g, taus, dy2,
                                        yoff=yoff,
                                        iron=IRON_UP)
        yoff -= ag

    tooth = TOOTH_ONLY if model_type in ('S1R2', 'S1R2_all') else TOOTH_DOWN
    yoff = _draw_horizontal_slots(ax,
                                  slot_height, slot_width, yoke_height,
                                  Q, g, taus,
                                  yoff=yoff,
                                  tooth=tooth)
    yoff -= ag

    iron = IRON_DOWN if model_type in ('S1R1', 'S1R2', 'S1R2_all') else IRON_NO
    yoff = _draw_horizontal_magnets(ax, poles,
                                    magn_height, rel_magn_width,
                                    magn_yoke_height,
                                    Q, g, taus, dy2,
                                    yoff=yoff,
                                    iron=iron)
    yoff -= ag

    if model_type in ('S2R1', 'S2R1_all'):  # 2 rotor
        yoff = _draw_horizontal_slots(ax,
                                      slot_height, slot_width, yoke_height,
                                      Q, g, taus,
                                      yoff=yoff,
                                      tooth=TOOTH_UP)

    _set_plot_attributes(ax)
    logger.debug("end of horizontal_plot()")


class AFPM:
    """Axial Flux PM
    Arguments:
    workdir: (str) working directory
    magnetizingCurves: (str) or (object)
    magnetMat: magnet material
    condMat: conductor material
    """
    def __init__(self, workdir, magnetizingCurves='.', magnetMat='',
                 condMat='', cmd=None):
        self.parstudy = parstudy.List(
            workdir, condMat=condMat, magnets=magnetMat,
            magnetizingCurves=magnetizingCurves, cmd=cmd)

    def __call__(self, engine, machine, simulation, num_slices):
        """ run FE simulation

        returns
        pos: (list of float)
        r1: radius
        torque: (list of float)
        emf: (list of float)
        emf_amp: (float)
        emf_angle: (float)
        freq: (float)
        currents: (list of list of float)
        plfe, plmag, plcu: (list of float)
        psid, psiq, psim: (float)
        i1, beta: (float)
        id, iq: (float)
        Ld, Lq: (float)
        """
        try:
            if machine['afmtype'] not in AFM_TYPES:
                raise ValueError(f"invalid afm type {machine['afmtype']}")
        except KeyError:
            raise ValueError("missing key afmtype")
        slotmodel = magnet_template_name(machine)
        logger.info("num_slices %d rmagw %s", num_slices,
                    machine['magnet'][slotmodel]['rel_magn_width'])
        if np.isscalar(machine['magnet'][slotmodel]['rel_magn_width']):
            rmagw = num_slices*[machine['magnet'][slotmodel]['rel_magn_width']]
        elif len(machine['magnet'][slotmodel]['rel_magn_width']) == 1: 
                rmagw = num_slices*machine['magnet'][slotmodel]['rel_magn_width']
        else: 
            rmagw = machine['magnet'][slotmodel]['rel_magn_width']
            num_slices = len(rmagw)

        lfe = get_arm_lengths(machine['outer_diam'],
                              machine['inner_diam'],
                              num_slices)
        pole_width = get_pole_widths(machine['outer_diam'],
                                     machine['inner_diam'],
                                     machine['poles'],
                                     num_slices)
        linspeed = [simulation['speed']*machine['poles']*pw
                    for pw in pole_width]

        if "num_agnodes" not in machine:
            Q1 = machine['stator']['num_slots']
            p = machine['poles']
            for pw in pole_width:
                machine['num_agnodes'] = num_agnodes(Q1, p//2, pw,
                                                     machine['airgap'])
            logger.info("Num agnodes/pole %d", machine['num_agnodes'])
        parvardef = {
            "decision_vars": [
                {"values": pole_width,
                 "name": "pole_width"},
                {"values": lfe,
                 "name": "lfe"},
                {"values": rmagw,
                 "name": f"magnet.{slotmodel}.rel_magn_width"},
                {"values": linspeed, "name": "speed"}
            ]
        }

        machine['pole_width'] = np.pi * machine['inner_diam']/machine['poles']
        machine['lfe'] = machine['outer_diam'] - machine['inner_diam']
        try:
            machine['magnet'][slotmodel]['rel_magn_width'] = max(
                machine['magnet'][slotmodel]['rel_magn_width'])
        except TypeError:
            pass

        simulation['skew_displ'] = (simulation.get('skew_angle', 0)/180 * np.pi
                                    * machine['inner_diam'])
        nlresults = {}
        if (simulation['calculationMode'] == 'torq_calc'
            and 'poc' not in simulation):
            nlcalc = dict(
                calculationMode="cogg_calc",
                magn_temp=simulation.get('magn_temp', 20),
                num_move_steps=60,
                skew_linear=0,
                skew_steps=0,
                poc=poc.Poc(machine['pole_width']),
                speed=0)
            logging.info("Noload simulation")
            nlresults = self.parstudy(parvardef,
                                      machine, nlcalc, engine)
            if nlresults['status'].count('C') != len(nlresults['status']):
                raise ValueError(f'Noload simulation failed {nlresults["status"]}')
            nlresults.update(process(lfe, pole_width, machine, nlresults['f']))

            current_angles = nlresults['f'][0]['current_angles']
            simulation['poc'] = poc.Poc(
                999,
                parameters={
                    'phi_voltage_winding': current_angles})
            logger.info("Current angles: %s", current_angles)
            simulation['calc_noload'] = 0

        if 'poc' not in simulation:
            simulation['poc'] = poc.Poc(machine['pole_width'])

        self.parstudy.result_func = get_magdata
        logger.info("Simulation %s", simulation['calculationMode'])
        lresults = self.parstudy(
            parvardef, machine, simulation, engine)
        if lresults['status'].count('C') != len(lresults['status']):
            raise ValueError(
                f'{simulation["calculationMode"]} failed {lresults["status"]}')

        results = process(lfe, pole_width, machine, lresults['f'])
        if nlresults:
            results.update(_psidq_ldq(results, nlresults))
        return results



"""Wound-rotor synchronous machine (EESM) electrical circuit model

"""
import logging
import warnings
import pathlib
import numpy as np
import scipy.optimize as so
import scipy.interpolate as ip
from .utils import skin_resistance, wdg_resistance, betai1, iqd, KTH, create_wdg, loss_models
from .. import parstudy, windings
import femagtools.bch

EPS = 1e-13

eecdefaults = {
    'zeta1': 0.3,
    'zeta2': 0,
    'gam': 0.7,
    'kh': 2,
    'tcu1': 20,
    'tcu2': 20,
    'rotor_mass': 0,
    'kfric_b': 1,
    'kpfe': 1 # common iron loss factor
}

logger = logging.getLogger('sm')
logging.captureWarnings(True)


def parident(workdir, engine, machine,
             magnetizingCurves, condMat,
             **kwargs):
    """return dict of parameters of equivalent circuit for
        electrically excited synchronous machines

        arguments:
        machine: dict() with machine parameters
        magnetizingCurves: list of dict() with BH curves
        condMat: list of dict() with conductor material properties

        optional arguments:
        num_cur_steps: number of current steps (default 5)
        num_beta_steps: number of beta steps (default 13)
        num_exc_steps: number of excitation current (default 6)
        speed: rotor speed in 1/s (default 160/p)
        i1_max: maximum current in A rms (default approx 3*i1nom)
        beta_min: minimal current angle (default -180°)
        beta_max: maximal current angle (default 0°)
        num_move_steps: number of move steps
        cmd: femag command (default None, platform executable)
        feloss: jordan, steinmetz, modified steinmetz, bertotti
        """
    cmd = kwargs.get('cmd', None)

    wdgk = 'windings' if 'windings' in machine else 'winding'
    g = machine[wdgk].get('num_par_wdgs', 1)
    N = machine[wdgk]['num_wires']
    if 'cufilfact' in machine[wdgk]:
        fcu = machine[wdgk]['cufilfact']
    elif 'fillfac' in machine[wdgk]:
        fcu = machine[wdgk]['fillfac']
    else:
        fcu = 0.42
    try: # calc basic dimensions if not fsl or dxf model
        from ..model import MachineModel
        wdg = create_wdg(machine)
        Q1 = wdg.Q
        model = MachineModel(machine)
        Jmax = 20e6  # max current density in A/m2
        Acu = fcu*model.slot_area()  # approx. copper area of one slot
        i1_max = round(g*Acu/wdg.l/N*Jmax/10)*10
    except KeyError:
        if kwargs.get('i1_max', 0) == 0:
            raise ValueError('i1_max missing')
        i1_max = kwargs['i1_max']


    if "magnetFsl" in machine["magnet"]:
        rotorkey = "magnet"
    else:
        rotorkey = "rotor"

    ifnom = machine[rotorkey]['ifnom']
    exc_logspace = True
    ifmin, ifmax = ifnom/4, 1.4*ifnom
    if exc_logspace:
        """excur = np.logspace(np.log(ifmin), np.log(ifmax),
                            kwargs.get("num_exc_steps", 6),
                            base=np.exp(1)).tolist()"""
        # create a data grid always contains ifnom
        excur = np.logspace(np.log(ifmin), np.log(ifnom),
                            kwargs.get("num_exc_steps", 6) - 2,
                            base=np.exp(1), endpoint=False).tolist()
        excur.extend([ifnom, ifmax])
    else:
        excur = np.linspace(ifmin, ifmax,
                            kwargs.get("num_exc_steps", 6))

    logger.info("Exc current %s", excur)
    parvardef = {
        "decision_vars": [
            {"values": excur, "name": "load_ex_cur"}
        ]
    }

    parvar = parstudy.List(
        workdir,  condMat=condMat,
        magnetizingCurves=magnetizingCurves, cmd=cmd)

    simulation = dict(
        calculationMode=kwargs.get('calculationMode',
                                   'ld_lq_fast'),
        wind_temp=20.0,
        i1_max=kwargs.get('i1_max', i1_max),
        maxid=kwargs.get('maxid', 0),
        minid=kwargs.get('minid', -i1_max),
        maxiq=kwargs.get('maxiq', i1_max),
        miniq=kwargs.get('miniq', -i1_max),
        delta_id=kwargs.get('delta_id', i1_max/5),
        delta_iq=kwargs.get('delta_iq', i1_max/5),
        beta_min=kwargs.get('beta_min', -180),
        beta_max=kwargs.get('beta_max', 0),
        num_move_steps=kwargs.get('num_move_steps', 31),
        load_ex_cur=0.5,
        num_cur_steps=kwargs.get('num_cur_steps', 5),
        num_beta_steps=kwargs.get('num_beta_steps', 13),
        num_par_wdgs=machine[wdgk].get('num_par_wdgs', 1),
        skew_angle=kwargs.get('skew_angle', 0.0),
        num_skew_steps=kwargs.get('num_skew_steps', 0.0),
        period_frac=kwargs.get('period_frac', 6),
        speed=kwargs.get('speed', 50))

    if kwargs.get("feloss", 0):
        simulation["feloss"] = kwargs["feloss"]
        machine["calc_fe_loss"] = loss_models[kwargs["feloss"].lower()]
    ###self.cleanup()  # remove previously created files in workdir
    results = parvar(parvardef, machine, simulation, engine)
    b = results['f'][-1]

    if 'poles' not in machine:  # dxf model?
        machine['poles'] = 2*results['f'][0]['machine']['p']
        da1 = 2*results['f'][0]['machine']['fc_radius']
        wdg = create_wdg(machine)

    if simulation['calculationMode'] == 'ld_lq_fast':
        idname = 'ldq'
    else:
        idname = 'psidq'
    for r in results['f']:
        for k in ('hf', 'ef'):
            if k in r['lossPar']:
                r[idname]['losses'][k] = r['lossPar'][k]
    try:
        rotor_mass = sum(results['f'][-1]['weights'][-1])
    except KeyError:
        rotor_mass = 0  # need femag classic > rel-9.3.x-48-gca42bbd0

    losskeys = ('speed', 'ef', 'hf', 'cf',
                'styoke_hyst', 'stteeth_hyst', 'styoke_eddy',
                'stteeth_eddy', 'rotor_hyst', 'rotor_eddy',
                'styoke_excess', 'stteeth_excess', 'rotor_excess')

    # winding resistance
    try:
        r1 = machine[wdgk]['resistance']
    except KeyError:
        try:
            Q1 = machine['stator']['num_slots']
            yd = machine[wdgk].get('coil_span', Q1/machine['poles'])
            wdg = windings.Winding(
                {'Q': Q1,
                 'm': machine[wdgk]['num_phases'],
                 'p': machine['poles']//2,
                 'l': machine[wdgk]['num_layers'],
                 'yd': yd})

            lfe = machine['lfe']
            da1 = machine['outer_diam']
            if 'dia_wire' in machine[wdgk]:
                aw = np.pi*machine[wdgk].get('dia_wire', 1e-3)**2/4
            else:  # wire diameter from slot area
                aw = 0.75 * fcu * np.pi*da1*hs/Q1/wdg.l/N
            r1 = wdg_resistance(wdg, N, g, aw, da1, hs, lfe)
        except (KeyError, NameError):
            from .. import nc
            model = nc.read(str(pathlib.Path(workdir) / machine['name']))
            try:
                nlayers = wdg.l
            except UnboundLocalError:
                wdg = create_wdg(machine)
                nlayers = wdg.l
                da1 = 2*results['f'][0]['machine']['fc_radius']
            Q1 = wdg.Q
            istat = 0 if model.get_areas()[0]['slots'] else 1
            asl = model.get_areas()[istat]['slots']
            # diameter of wires
            aw = fcu*asl/Q1/nlayers/N
            hs = asl/(np.pi*da1/3)
            r1 = wdg_resistance(wdg, N, g, aw, da1, hs, lfe)

    if simulation['calculationMode'] == 'ld_lq_fast':
        dqpars = dict(m=3, p=b['machine']['p'],
                      r1=float(r1),
                      r2=machine[rotorkey].get('resistance', 1),
                      rotor_mass=rotor_mass, kfric_b=1,
                      ldq=[dict(
                          ex_current=b['machine']['ex_current'],
                          i1=b['ldq']['i1'],
                          beta=b['ldq']['beta'],
                          psid=b['ldq']['psid'],
                          psiq=b['ldq']['psiq'],
                          torque=b['ldq']['torque'],
                          ld=b['ldq']['ld'],
                          lq=b['ldq']['lq'],
                          losses={k: b['ldq']['losses'][k]
                                    for k in losskeys if k in b['ldq']['losses']})
                         for b in results['f']])
    else:
        dqpars = dict(m=3, p=b['machine']['p'],
                      r1=r1,
                      r2=machine[rotorkey].get('resistance', 1),
                      rotor_mass=rotor_mass, kfric_b=1,
                      psidq=[dict(
                          ex_current=b['machine']['ex_current'],
                          iq=b['psidq']['iq'],
                          id=b['psidq']['id'],
                          psidq_ldq=b['psidq_ldq'],
                          psid=b['psidq']['psid'],
                          psiq=b['psidq']['psiq'],
                          torque=b['psidq']['torque'],
                          losses={k: b['psidq']['losses'][k]
                                for k in losskeys if k in b['psidq']['losses']})
                       for b in results['f']])

    if 'current_angles' in results['f'][0]:
        dqpars['current_angles'] = results['f'][0]['current_angles']
    return dqpars

def _linsampl(exc, excl, a):
    """auxiliary func for linear sampling of nonlinear sequence
        arguments:
        exc: (list) nonlinear sequence of excitation current
        excl: (list) linear sequence of excitation current
        a: (array) matrix to be resampled"""
    z = []
    b, i = a.shape[1:]
    for x in range(b):
        for y in range(i):
            zl = ip.interp1d(exc, a[:, x, y], kind='linear')
            z.append(zl(excl))
    return np.array(z).T.reshape(len(excl), b, i)


def _splinterp(beta, i1, betax, i1x, a):
    """auxiliary function to increase resolution of array a
        using a cubic spline interpolation.
        arguments:
        beta: (list) n original up-i angles in rad
        i1: (list) m original current in A
        betax: (list) nx new up-i angles in rad
        i1x: (list) mx new currents
        a: (nxm array) to be interpolated"""
    f = ip.RectBivariateSpline(
        beta, i1, np.asarray(a),
        kx=3, ky=3).ev
    X, Y = np.meshgrid(betax, i1x)
    return f(X, Y).T


def _islinear(exc):
    d = np.diff(exc)
    m = np.mean(d)
    return np.max(np.abs(d**2 - m**2)) < 1e-9


def _gradient_respecting_bounds(bounds, fun, eps=1e-8):
    """bounds: list of tuples (lower, upper)"""
    def gradient(x):
        logger.info(x)
        fx = fun(x)
        grad = np.zeros(len(x))
        for k in range(len(x)):
            d = np.zeros(len(x))
            d[k] = eps if x[k] + eps <= bounds[k][1] else -eps
            grad[k] = (fun(x + d) - fx) / d[k]
        return grad
    return gradient


class SynchronousMachine(object):
    """ represent Synchronous machine with wound rotor (EESM)

    Arguments:
        eecpars: dict() electrical circuit parameters
        """
    def __init__(self, eecpars, **kwargs):
        self.kth1 = KTH
        self.kth2 = KTH
        self.skin_resistance = [None, None]
        self.kpmag = 1
        self.bertotti = False
        # here you can set user defined functions for calculating the skin-resistance,
        # according to the current frequency w. First function in list is for stator, second for rotor.
        # If None, the femagtools intern default implementation is used.
        # User defined functions need to have the following arguments:
        # - r0: (float) dc-resistance at 20°C
        # - w: (float)  current frequency in rad (2*pi*f)
        # - tcu: (float) conductor temperature in deg Celsius
        # - kth: (float) temperature coefficient (Default = 0.0039, Cu)
        for k in eecdefaults.keys():
            setattr(self, k, eecdefaults[k])

        for k in eecpars:
            if k not in ('ldq', 'psidq'):
                setattr(self, k, eecpars[k])

        for k in kwargs:
            setattr(self, k, kwargs[k])

        try:
            self.tfric = self.kfric_b*self.rotor_mass*30e-3/np.pi
        except AttributeError:
            self.tfric = 0

        self.fo = 50
        self.plexp = {'styoke_hyst': [1.0, 1.0],
                      'stteeth_hyst': [1.0, 1.0],
                      'styoke_eddy': [2.0,2.0],
                      'stteeth_eddy': [2.0,2.0],
                      'rotor_hyst': [1.0,1.0],
                      'rotor_eddy': [2.0, 2.0]}

    def _set_losspar(self, speed, ef, hf):
        self.fo = speed*self.p
        self.plexp = {'styoke_hyst': hf,
                      'stteeth_hyst': hf,
                      'styoke_eddy': ef,
                      'stteeth_eddy': ef,
                      'rotor_hyst': hf,
                      'rotor_eddy': ef}

        if self.bertotti:
            self.plexp.update({
                'styoke_excess': 1.5,
                'stteeth_excess':1.5,
                'rotor_excess': 1.5})

    def pfric(self, n):
        """friction and windage losses"""
        return 2*np.pi*n*self.tfric

    def rstat(self, w):
        """stator resistance"""
        if isinstance(self.zeta1, list):
            # polyfit from ac loss calculation
            freq = w/2/np.pi
            kr = self.zeta1[0]*freq**3 + self.zeta1[1]*freq**2 + \
                self.zeta1[2]*freq + self.zeta1[3]
            if isinstance(kr, list):
                kr = np.array(kr)
                kr[kr<1.0] = 1.0
            elif isinstance(kr, np.ndarray):
                kr[kr<1.0] = 1.0
            else:
                if kr < 1.0:
                    kr = 1.0
            return self.r1*(1 + self.kth1*(self.tcu1 - 20))*kr  # ref 20°C
        sr = self.skin_resistance[0]
        if sr is not None:
            return sr(self.r1, w, self.tcu1, kth=self.kth1)
        else:
            return skin_resistance(self.r1, w, self.tcu1, self.zeta1,
                                   self.gam, self.kh, kth=self.kth1)

    def rrot(self, w):
        """rotor resistance"""
        sr = self.skin_resistance[1]
        if sr is not None:
            return sr(self.r2, w, self.tcu2, kth=self.kth2)
        else:
            return skin_resistance(self.r2, w, self.tcu2, self.zeta2,
                                   0.0, 1, kth=self.kth2)

    def torque_iqd(self, iq, id, iex):
        "torque at q-d-current"
        psid, psiq = self.psi(iq, id, iex)
        return self.m*self.p/2*(psid*iq - psiq*id)

    def tloss_iqd(self, iq, id, iex, n):
        """return loss torque of d-q current, iron loss correction factor
            and friction windage losses"""
        if n > 1e-3:
            f1 = self.p*n
            plfe1 = self.kpfe * self.iqd_plfe1(iq, id, iex, f1)
            plfe2 = self.kpfe * self.iqd_plfe2(iq, id, iex, f1)
            if hasattr(self, 'kpfe_s'):
                plfe1 = self.kpfe_s * self.iqd_plfe1(iq, id, iex, f1)
            if hasattr(self, 'kpfe_r'):
                plfe2 = self.kpfe_r * self.iqd_plfe2(iq, id, iex, f1)
            plfe = plfe1 + plfe2
            return (plfe + self.pfric(n))/(2*np.pi*n)
        return 0

    def tmech_iqd(self, iq, id, iex, n):
        """return shaft torque of d-q current and speed"""
        return self.torque_iqd(iq, id, iex) - self.tloss_iqd(iq, id, iex, n)

    def torquemax(self, i1, iex):
        "returns maximum torque of i1, iex (nan if i1 out of range)"
        def torquei1b(b):
            return self.torque_iqd(*iqd(b[0], i1), iex)
        res = so.minimize(torquei1b, (0,))
        return res.fun

    def torquemin(self, i1, iex):
        "returns minimum torque of i1, iex (nan if i1 out of range)"
        def torquei1b(b):
            return self.torque_iqd(*iqd(b[0], i1), iex)
        res = so.minimize(torquei1b, (-np.pi/2,))
        return -res.fun

    def uqd(self, w1, iq, id, iex):
        """return uq, ud of frequency w1 and d-q current"""
        psid, psiq = self.psi(iq, id, iex)
        # r1 = self.rstat(w1)
        return self.r1*iq + w1*(self.ls*id + psid), self.r1*id - w1*(self.ls*iq + psiq)

    def plcu1(self, iqde, w1):
        r1 = self.rstat(w1)
        return 3/2*r1*(iqde[0]**2 + iqde[1]**2)

    def plcu2(self, iqde, w1=0):
        r2 = self.rrot(0)
        return r2*iqde[2]**2

    def culoss(self, iqde, w1=0):
        return self.plcu1(iqde, w1) + self.plcu2(iqde, 0)

    def iqd_plcu1(self, iq, id, f1):
        return self.plcu1((iq, id), 2*np.pi*f1)

    def iqd_plcu2(self, iq, id, iex):
        return self.plcu2((iq, id, iex))

    def iqd_plfe2_(self, iq, id, f1):
        return np.zeros(np.asarray(iq).shape)

    def iqd_plmag(self, iq, id, f1):
        return np.zeros(np.asarray(iq).shape)

    def iqd_tmech(self, torque, n, disp=False, maxiter=500):
        """return currents for shaft torque with minimal losses"""
        if torque > 0:
            startvals = self.bounds[0][1]/2, 0, self.bounds[-1][1]
        else:
            startvals = -self.bounds[0][1]/2, 0, self.bounds[-1][1]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            res = so.minimize(
                self.culoss, startvals, method='SLSQP',  # trust-constr
                bounds=self.bounds,
                #            jac=gradient_respecting_bounds(self.bounds, self.culoss),
                constraints=[
                    {'type': 'eq',
                     'fun': lambda iqd: self.tmech_iqd(*iqd, n) - torque}])
            #options={'disp': disp, 'maxiter': maxiter})
            if res['success']:
                return res.x
            else:
                res = so.minimize(
                    lambda cur: (self.tmech_iqd(*cur, n) - torque)**2, res.x, method='SLSQP',
                    bounds=self.bounds,
                )
                if res['success']:
                    return res.x

        logger.warning("%s: torque=%f %f, io=%s",
                       res['message'], torque, self.tmech_iqd(*startvals, n),
                       startvals)
        raise ValueError(res['message'])

    def iqd_torque(self, torque, disp=False, maxiter=500):
        """return currents for torque with minimal losses"""
        if torque > 0:
            startvals = self.bounds[0][1]/2, 0, self.bounds[-1][1]
        else:
            startvals = -self.bounds[0][1]/2, 0, self.bounds[-1][1]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            res = so.minimize(
                self.culoss, startvals, method='SLSQP',  # trust-constr
                bounds=self.bounds,
                #            jac=gradient_respecting_bounds(self.bounds, self.culoss),
                constraints=[
                    {'type': 'eq',
                     'fun': lambda iqd: self.torque_iqd(*iqd) - torque}])
            #options={'disp': disp, 'maxiter': maxiter})
            if res['success']:
                return res.x
        logger.warning("%s: torque=%f %f, io=%s",
                       res['message'], torque, self.torque_iqd(*startvals),
                       startvals)
        raise ValueError(res['message'])

    def mtpa(self, i1max):
        """return iq, id, iex currents and maximum torque per current """
        T0 = self.torque_iqd(np.sqrt(2)*i1max, 0, self.bounds[-1][1])
        def i1tq(tq):
            return abs(i1max) - np.linalg.norm(self.iqd_torque(tq)[:2])/np.sqrt(2)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tq = so.fsolve(i1tq, T0)[0]
            iq, id, iex = self.iqd_torque(tq)
        return iq, id, iex, tq

    def mtpa_tmech(self, i1max, n):
        """return iq, id, iex currents and maximum torque per current """
        T0 = self.torque_iqd(np.sqrt(2)*i1max, 0, self.bounds[-1][1])
        def i1tq(tq):
            return i1max - np.linalg.norm(self.iqd_tmech(tq, n)[:2])/np.sqrt(2)
        tq = so.fsolve(i1tq, T0)[0]
        iq, id, iex = self.iqd_tmech(tq, n)
        return iq, id, iex, tq

    def iqd_tmech_umax(self, torque, w1, u1max, log=0, **kwargs):
        """return currents and shaft torque at stator frequency and
         with minimal losses at max voltage"""
        iqde = self.iqd_tmech(torque, w1/2/np.pi/self.p)
        if np.linalg.norm(
            self.uqd(w1, *iqde)) <= u1max*np.sqrt(2):
            if log:
                log(iqde)
            return (*iqde, torque)
        beta, i1 = betai1(iqde[0], iqde[1])
        iex = iqde[2]
        def ubeta(b):
            return np.sqrt(2)*u1max - np.linalg.norm(
                         self.uqd(w1, *iqd(b, i1), iex))
        beta = -np.pi/4 if torque>0 else -3*np.pi/4
        io = *iqd(beta, i1), iex

        #    logger.debug("--- torque %g io %s", torque, io)
        n = w1/2/np.pi/self.p
        res = so.minimize(
            self.culoss, io, method='SLSQP',  # trust-constr
            bounds=self.bounds,
            constraints=[
                {'type': 'eq',
                 'fun': lambda iqd: self.tmech_iqd(*iqd, n) - torque},
                {'type': 'eq',
                 'fun': lambda iqd: np.linalg.norm(
                     self.uqd(w1, *iqd)) - u1max*np.sqrt(2)}])

        if log:
            log(res.x)
        if res["success"]:
            return *res.x, self.tmech_iqd(*res.x, n)
        else:
            # didn't converge, fullfill voltage and torque demand
            res = so.minimize(
                lambda cur: (self.tmech_iqd(*cur, n) - torque)**2, io, method='SLSQP',
                bounds=self.bounds,
                constraints=[
                {'type': 'eq',
                 'fun': lambda iqd: np.linalg.norm(
                     self.uqd(w1, *iqd)) - u1max*np.sqrt(2)}]
            )
            if res["success"]:
                return *res.x, self.tmech_iqd(*res.x, n)
            else:
                res = so.minimize(
                    lambda cur: (self.tmech_iqd(*cur, n) - torque)**2 +
                    (np.linalg.norm(self.uqd(w1, *cur)) - u1max*np.sqrt(2))**2, io, method='SLSQP',
                    bounds=self.bounds,
                )
                return *res.x, self.tmech_iqd(*res.x, n)

        # logger.warning("%s: w1=%f torque=%f, u1max=%f, io=%s",
        #                res['message'], w1, torque, u1max, io)
        # raise ValueError(res['message'])

    def iqd_torque_umax(self, torque, w1, u1max,
                        disp=False, maxiter=500, log=0, **kwargs):
        """return currents for torque with minimal losses"""
        iqde = self.iqd_torque(torque, disp, maxiter)
        if np.linalg.norm(
                self.uqd(w1, *iqde)) <= u1max*np.sqrt(2):
                if log:
                    log(iqde)
                return (*iqde, torque)
        beta, i1 = betai1(iqde[0], iqde[1])
        iex = iqde[2]
        def ubeta(b):
            return np.sqrt(2)*u1max - np.linalg.norm(
                         self.uqd(w1, *iqd(b, i1), iex))
        beta = -np.pi/4 if torque>0 else -3*np.pi/4
        io = *iqd(beta, i1), iex
        res = so.minimize(
            self.culoss, io, method='SLSQP',  # trust-constr
            bounds=self.bounds,
            #options={'disp': disp, 'maxiter': maxiter},
            #            jac=gradient_respecting_bounds(self.bounds, self.culoss),
            constraints=[
                {'type': 'eq',
                 'fun': lambda iqd: self.torque_iqd(*iqd) - torque},
                {'type': 'eq',
                 'fun': lambda iqd: u1max*np.sqrt(2) - np.linalg.norm(
                     self.uqd(w1, *iqd))}])
        if log:
            log(res.x)
        if res['success']:
            return *res.x, self.torque_iqd(*res.x)
        logger.warning("%s: w1=%f torque=%f, u1max=%f, io=%s",
                       res['message'], w1, torque, u1max, io)
        raise ValueError(res['message'])

    def w1_imax_umax(self, i1max, u1max):
        """return frequency w1 and shaft torque at voltage u1max and current i1max

        Keyword arguments:
            u1max -- the maximum voltage (Vrms)
            i1max -- the maximum current (Arms)"""
        iq, id, iex, T = self.mtpa(i1max)
        n0 = np.sqrt(2)*u1max/np.linalg.norm(
            self.psi(iq, id, iex))/2/np.pi/self.p
        return self.w1_umax(u1max, iq, id, iex), T

    def w1_umax(self, u, iq, id, iex):
        """return frequency w1 at given voltage u and id, iq current

        Keyword arguments:
            u -- the maximum voltage (RMS)
            iq, id -- the d-q currents"""
        w10 = np.sqrt(2)*u/np.linalg.norm(self.psi(iq, id, iex))
        return so.fsolve(
            lambda w1: np.linalg.norm(self.uqd(w1, iq, id, iex))-u*np.sqrt(2),
            w10)[0]

    def characteristics(self, T, n, u1max, nsamples=50,
                        with_tmech=True, with_torque_corr=False, **kwargs):
        """calculate torque speed characteristics.
        return dict with list values of
        n, T, u1, i1, beta, cosphi, pmech, n_type

        Keyword arguments:
        T -- (float) the maximum torque in Nm
        n -- (float) the maximum speed in 1/s
        u1max -- (float) the maximum voltage in V rms
        nsamples -- (optional) number of speed samples
        with_tmech -- (optional) use friction and windage losses
        with_torque_corr -- (optional) T is corrected if out of range
        Optional arguments:
        i1max: max. phase current (RMS)
        """
        if kwargs.get('i1max', 0):
            w1type, T = self.w1_imax_umax(kwargs['i1max'], u1max)

        try:
            iq, id, iex = self.iqd_torque(T)
        except ValueError:
            tmax = self.torquemax(
                    self.i1range[1], self.exc_max)
            tmin = 0
            if self.betarange[0] < -np.pi/2:
                tmin = -self.torquemin(
                        self.i1range[1], self.exc_max)
            if with_torque_corr:
                Torig = T
                if T > 0:
                    T = np.floor(0.94*tmax)
                else:
                    T = np.ceil(0.94*tmin)
                logger.warning("corrected torque %f -> %f Nm",
                               Torig, T)
                iq, id, iex = self.iqd_torque(T)
            else:
                raise ValueError(
                        f"torque {T} Nm out of range ({tmin:.1f}, {tmax:.1f} Nm)")

        if with_tmech:
            i1max = betai1(iq, id)[1]
            if T < 0:
                i1max = -i1max
            w1type, Tf = self.w1_imax_umax(i1max, u1max)

        else:
            Tf = T
            w1type = self.w1_umax(u1max, iq, id, iex)
        logger.debug("w1type %f", w1type)
        wmType = float(w1type/self.p)
        pmax = Tf*wmType

        def tload(wm):
            if abs(wm*Tf) < abs(pmax):
                return Tf
            return pmax/wm

        wmtab = []
        dw = 0
        wmMax = 2*np.pi*n
        '''
        wmMax = 5*wmType
        if n > 0:
            wmMax = min(wmMax, 2*np.pi*n)
        '''
        if wmType > wmMax:
            wmrange = sorted([0, wmMax])
            wmtab = np.linspace(0, wmMax, nsamples).tolist()
        else:
            wmrange = sorted([0, wmType, wmMax])
            nx = int(round(nsamples*wmType/wmMax))
            dw = (wmMax-wmType)/(nsamples-nx)
            wmtab = (np.linspace(0, wmType, nx).tolist() +
                     np.linspace(wmType+dw, wmMax, nsamples-nx).tolist())

        logger.info("Speed range T %f %s", Tf, wmrange)
        wmtab[0] = 0

        r = dict(u1=[], i1=[], id=[], iq=[], iex=[], T=[], cosphi=[], n=[],
                 beta=[], plfe1=[], plfe2=[], plcu1=[], plcu2=[])
        # add type speed to result dict
        r['n_type'] = wmType/2/np.pi
        for wm, tq in zip(wmtab, [tload(wx) for wx in wmtab]):
            w1 = wm*self.p
            if wm > wmType:
                if with_tmech:
                    iq, id, iex, tqx = self.iqd_tmech_umax(
                            tq, w1, u1max)
                else:
                    iq, id, iex, tqx = self.iqd_torque_umax(
                            tq, w1, u1max)
                    tqx -= self.tfric
            else:
                tqx = Tf

            uq, ud = self.uqd(w1, iq, id, iex)
            u1 = np.linalg.norm((uq, ud))/np.sqrt(2)
            f1 = w1/2/np.pi
            r['id'].append(id)
            r['iq'].append(iq)
            r['iex'].append(iex)
            r['u1'].append(u1)
            beta = np.arctan2(id, iq)
            if beta > 0:
                beta -= 2*np.pi
            i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
            r['i1'].append(i1)
            r['beta'].append(beta/np.pi*180)
            gamma = np.arctan2(ud, uq)
            r['cosphi'].append(np.cos(gamma - beta))
            r['plfe1'].append(self.iqd_plfe1(iq, id, iex, f1))
            r['plfe2'].append(self.iqd_plfe2(iq, id, iex, f1))
            r['plcu1'].append(self.m*i1**2*self.rstat(w1))
            r['plcu2'].append(iex**2*self.rrot(0))
            r['T'].append(tqx)
            r['n'].append(wm/2/np.pi)

        r['plfe'] = (np.array(r['plfe1']) + np.array(r['plfe2'])).tolist()
        r['plcu'] = (np.array(r['plcu1']) + np.array(r['plcu2'])).tolist()
        r['plfw'] = [self.pfric(n) for n in r['n']]
        r['pmech'] = [2*np.pi*n*tq
                      for n, tq in zip(r['n'], r['T'])]
        pmech = np.array(r['pmech'])
        pltotal = (np.array(r['plfe1']) + np.array(r['plfw']) +
                   np.array(r['plcu1']) + np.array(r['plcu2']))
        r['losses'] = pltotal.tolist()

        if pmech.any():
            p1 = pmech + pltotal
            if np.abs(pmech[0]) < 1e-12:
                r['eta'] = [0]
                i = 1
            else:
                r['eta'] = []
                i = 0
            if np.all(abs(p1[i:]) > abs(pmech[i:])):
                r['eta'] += (pmech[i:]/(p1[i:])).tolist()
            else:
                r['eta'] += (p1[i:]/pmech[i:]).tolist()

        return r

class SynchronousMachinePsidq(SynchronousMachine):

    def __init__(self, eecpars, lfe=1, wdg=1, **kwargs):
        super(self.__class__, self).__init__(
            eecpars, **kwargs)
        self.iqrange = (eecpars['psidq'][0]['iq'][0],
                        eecpars['psidq'][0]['iq'][-1])
        self.idrange = (eecpars['psidq'][0]['id'][0],
                        eecpars['psidq'][0]['id'][-1])

        self.betarange = (-np.pi if min(self.iqrange) < 0 else -np.pi/2,
                          0 if max(self.iqrange) > 0 else -np.pi/2)
        self.i1range = (0, betai1(np.max(self.iqrange), 0)[1])
        islinear = True
        iexc = [l['ex_current'] for l in eecpars['psidq']]
        self.exc_max = iexc[-1]
        id = [i/wdg for i in eecpars['psidq'][-1]['id']]
        idx = np.linspace(id[0], id[-1], 12)
        iq = [i/wdg for i in eecpars['psidq'][-1]['iq']]
        iqx = np.linspace(iq[0], iq[-1], 20)

        if _islinear(iexc):
            exc = iexc
            psid = wdg*lfe*np.array([
                _splinterp(iq, id, iqx, idx, l['psid'])
                for l in eecpars['psidq']])
            psiq = wdg*lfe*np.array([
                _splinterp(iq, id, iqx, idx, l['psiq'])
                for l in eecpars['psidq']])
        else:
            islinear = False
            nsamples = 10
            iexcl = np.linspace(iexc[0], iexc[-1], nsamples)
            exc = iexcl
            psid = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                [_splinterp(iq, id, iqx, idx, l['psid'])
                 for l in eecpars['psidq']]))
            psiq = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                [_splinterp(iq, id, iqx, idx, l['psiq'])
                 for l in eecpars['psidq']]))

        self.psidf = ip.RegularGridInterpolator(
            (exc, iqx, idx), psid,
            method='cubic', bounds_error=False, fill_value=None)
        self.psiqf = ip.RegularGridInterpolator(
            (exc, iqx, idx), psiq,
            method='cubic', bounds_error=False, fill_value=None)
        # excitation currents bounds should respect ifnom
        self.bounds = [(min(iq), max(iq)),
                       (min(id), 0),
                       (iexc[0], iexc[-1])]
        # iron losses
        idname = 'psidq'
        keys = [k for k in self.plexp.keys() if k in eecpars[idname][0]['losses']]
        try:
            # check if bertotti
            if 'styoke_excess' in eecpars[idname][0]['losses'] and \
               np.any(np.array(eecpars[idname][0]['losses']['styoke_excess'])):
                self.bertotti = True
                keys += ['styoke_excess', 'stteeth_excess','rotor_excess']
            if islinear:
                pfe = {k: np.array([l['losses'][k]
                                    for l in eecpars[idname]])
                       for k in keys}
            else:
                pfe = {k: _linsampl(iexc, iexcl,
                                    np.array([l['losses'][k]
                                             for l in eecpars[idname]]))
                       for k in keys}
                self._losses = {k: ip.RegularGridInterpolator(
                    (exc, iq, id), lfe*pfe[k],
                    method='cubic', bounds_error=False, fill_value=None)
                            for k in keys}
                self._set_losspar(eecpars[idname][0]['losses']['speed'],
                                  eecpars[idname][0]['losses']['ef'],
                                  eecpars[idname][0]['losses']['hf'])
        except KeyError:
            logger.warning("loss map missing")
            self._losses = {k: lambda x: 0 for k in (
                'styoke_hyst', 'stteeth_hyst',
                'styoke_eddy', 'stteeth_eddy',
                'rotor_hyst', 'rotor_eddy')}

    def set_max_iex(self, iexc):
        self.bounds[-1] = (self.bounds[-1][0], iexc)
        self.exc_max = iexc

    def psi(self, iq, id, iex):
        """return psid, psiq of currents iq, id"""
        try:
            return self.psidf((iex, iq, id)), self.psiqf((iex, iq, id))
        except ValueError as ex:
            logger.error(iex, iq, id)
            raise ex

    def plfe1(self, iq, id, iex, f1):
        losskeys = [k for k in self._losses if k in [
            'styoke_eddy', 'styoke_hyst',
            'stteeth_eddy', 'stteeth_hyst']]
        if self.bertotti:
            losskeys += ['styoke_excess', 'stteeth_excess']
        return np.sum([
            self._losses[k]((iex, iq, id))*(f1/self.fo)**self.plexp[k][0]
            for k in losskeys], axis=0)

    def plfe2(self, iq, id, iex, f1):
        losskeys = ['rotor_hyst', 'rotor_eddy']
        if self.bertotti:
            losskeys += ['rotor_excess']
        return np.sum([
            self._losses[k]((iex, iq, id))*(f1/self.fo)**self.plexp[k][0]
            for k in losskeys], axis=0)

    def iqd_plfe1(self, iq, id, iex, f1):
        return self.plfe1(iq, id, iex, f1)

    def iqd_plfe2(self, iq, id, iex, f1):
        return self.plfe2(iq, id, iex, f1)

class SynchronousMachineLdq(SynchronousMachine):
    def __init__(self, eecpars, lfe=1, wdg=1, **kwargs):
        super(self.__class__, self).__init__(eecpars, **kwargs)
        self.betarange = (eecpars['ldq'][0]['beta'][0]/180*np.pi,
                          eecpars['ldq'][0]['beta'][-1]/180*np.pi)
        self.i1range = (0, eecpars['ldq'][0]['i1'][-1])
        islinear = True
        iexc = [l['ex_current'] for l in eecpars['ldq']]
        self.exc_max = iexc[-1]
        i1 = [i/wdg for i in eecpars['ldq'][-1]['i1']]
        i1x = np.linspace(i1[0], i1[-1], 12)
        beta = [b*np.pi/180 for b in eecpars['ldq'][-1]['beta']]
        betax = np.linspace(beta[0], beta[-1], 20)

        if _islinear(iexc):
            logger.info("Linear sampled ex current: %s",
                        iexc)
            exc = iexc
            psid = wdg*lfe*np.array([
                _splinterp(beta, i1, betax, i1x, l['psid'])
                    for l in eecpars['ldq']])
            psiq = wdg*lfe*np.array([
                _splinterp(beta, i1, betax, i1x, l['psiq'])
                    for l in eecpars['ldq']])
        else:
            islinear = False
            logger.info("Non Linear sampled ex current %s",
                        iexc)
            nsamples = 10
            iexcl = np.linspace(iexc[0], iexc[-1], nsamples)
            exc = iexcl
            psid = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                [_splinterp(beta, i1, betax, i1x, l['psid'])
                     for l in eecpars['ldq']]))
            psiq = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                [_splinterp(beta, i1, betax, i1x, l['psiq'])
                     for l in eecpars['ldq']]))

        # extrapolate outside range
        self.psidf = ip.RegularGridInterpolator(
            (exc, betax, i1x), np.sqrt(2)*psid,
            method='cubic',
            bounds_error=False, fill_value=None)
        self.psiqf = ip.RegularGridInterpolator(
            (exc, betax, i1x), np.sqrt(2)*psiq,
            method='cubic'
            , bounds_error=False, fill_value=None)
        i1max = np.sqrt(2)*(max(i1))
        # excitation currents bounds should respect ifnom
        self.bounds = [(np.cos(min(beta))*i1max, i1max),
                       (-i1max, 0),
                       (iexc[0], iexc[-1])]

        # iron losses
        idname = 'ldq'
        keys = [k for k in self.plexp.keys() if k in eecpars[idname][0]['losses']]
        try:
            # check bertotti losses
            if 'styoke_excess' in eecpars[idname][0]['losses'] and \
               np.any(np.array(eecpars[idname][0]['losses']['styoke_excess'])):
                self.bertotti = True
                keys += ['styoke_excess', 'stteeth_excess','rotor_excess']

            if islinear:
                pfe = {k: np.array([l['losses'][k]
                                    for l in eecpars[idname]])
                       for k in keys}
            else:
                pfe = {k: _linsampl(iexc, iexcl,
                                    np.array([l['losses'][k]
                                             for l in eecpars[idname]]))
                       for k in keys}

        # fill value with nan outside range
            self._losses = {k: ip.RegularGridInterpolator(
                (exc, beta, i1), lfe*pfe[k],
                method='cubic', bounds_error=False, fill_value=None)
                            for k in pfe.keys()}
            self._set_losspar(eecpars[idname][0]['losses']['speed'],
                              eecpars[idname][0]['losses']['ef'],
                              eecpars[idname][0]['losses']['hf'])
        except KeyError:
            logger.warning("loss map missing")
            self._losses = {k: lambda x: 0 for k in (
                'styoke_hyst', 'stteeth_hyst',
                'styoke_eddy', 'stteeth_eddy',
                'rotor_hyst', 'rotor_eddy')}
            pass

    def set_max_iex(self, iexc):
        self.bounds[-1] = (self.bounds[-1][0], iexc)
        self.exc_max = iexc

    def psi(self, iq, id, iex):
        """return psid, psiq of currents iq, id"""
        beta = np.arctan2(id, iq)
        if beta > 0:
            beta -= 2*np.pi
        i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
        try:
            return self.psidf((iex, beta, i1)), self.psiqf((iex, beta, i1))
        except ValueError as ex:
            logger.error("iex %s iq %f id %f beta %f i1 %f",
                         iex, iq, id, beta, i1)
            raise ex

    def plfe1(self, beta, i1, iex, f1):
        losskeys = ['styoke_eddy', 'styoke_hyst',
                     'stteeth_eddy', 'stteeth_hyst']
        if self.bertotti:
            losskeys += ['styoke_excess', 'stteeth_excess']
        return np.sum([
            self._losses[k]((iex, beta, i1))*(f1/self.fo)**self.plexp[k][0]
            for k in losskeys if k in self._losses], axis=0)

    def plfe2(self, beta, i1, iex, f1):
        losskeys =  ['rotor_hyst', 'rotor_eddy']
        if self.bertotti:
            losskeys += ['rotor_excess']
        return np.sum([
            self._losses[k]((iex, beta, i1))*(f1/self.fo)**self.plexp[k][0]
            for k in losskeys], axis=0)

    def iqd_plfe1(self, iq, id, iex, f1):
        beta = np.arctan2(id, iq)
        if np.isscalar(beta):
            if beta > 0:
                beta -= 2*np.pi
        else:
            beta[beta > 0] -= 2*np.pi
        i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
        return self.plfe1(beta, i1, iex, f1)

    def iqd_plfe2(self, iq, id, iex, f1):
        beta = np.arctan2(id, iq)
        if np.isscalar(beta):
            if beta > 0:
                beta -= 2*np.pi
        else:
            beta[beta > 0] -= 2*np.pi
        i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
        return self.plfe2(beta, i1, iex, f1)


if __name__ == '__main__':
    import sys
    import json
    import femagtools.plot
    import matplotlib.pyplot as plt
    with open(sys.argv[1]) as fp:
        eecpar = json.load(fp)
    if 'ldq' in eecpar:
        m = SynchronousMachineLdq(eecpar)
    else:
        m = SynchronousMachinePsidq(eecpar)
        T = 240
        u1max = 163
        nmax = 1000
        r = m.characteristics(T, 0, u1max)
        femagtools.plot.characteristics(r, 'SynchronousMachine')
        plt.show()



# -*- coding: utf-8 -*-
"""analyze windings

 Conventions

 Number of slots: Q
 Numper of pole pairs: p
 Number of phases: m
 Number of layers: l
 Number of wires per slot side: n
 Number of slots per pole and phase: q = Q/p/2/m
 Number of coils per phase: c = Q * l/2/m
 Number of parallel circuits (coil groups): g
 Number of turns per phase: w1 = Q * n * l/2/m/g

Refs:
odd number of phases
 doi:10.1007/s00502-021-00873-6
even number of phases
 doi:10.1049/iet-epa.2020.0553
"""
import numpy as np
import femagtools.bch
from xml.etree import ElementTree as ET

coil_color = ['lime', 'gold', 'magenta',
              'blue', 'brown', 'blueviolet']


def num_basic_windings(Q, p, l):
    """return number of basic windings"""
    if l == 1:  # single layer
        return np.gcd(Q//2, p)
    return np.gcd(Q, p)


def q1q2yk(Q, p, m, l=1):
    """returns q1, q2, Yk, Qb"""
    t = num_basic_windings(Q, p, l)
    Qb = Q//t
    qbb = Qb if l==2 else Qb//2
    pb = p//t
    if qbb//m % 2:  # odd
        q2 = (qbb + m)//(2*m) - 1
        q1 = q2 + 1
    else:
        q2 = (qbb)//(2*m)
        q1 = q2
    n = 1
    while (n*qbb + 1) % pb:
        n += 1
    Yk = (n*qbb + 1)//pb
    return q1, q2, Yk, Qb

def end_wdg_length_round_wires(layers, Rp, Rb, r_wire, h, coil_span, Q, bore_diam, slot_h1, slot_height):
    '''return length of a single winding head for 1 coil turn.
    Multiply by 2 to get length for both winding heads'''
    if layers == 2:
        R_inner_lyr = bore_diam/2 + slot_h1 + slot_height/4
        R_outer_lyr = bore_diam/2 + slot_h1 + 3*slot_height/4
    elif layers == 1:
        R_inner_lyr = bore_diam/2 + slot_h1 + slot_height/2
        R_outer_lyr = bore_diam/2 + slot_h1 + slot_height/2
    else:
        raise ValueError("Round wire windings can only have 1 or 2 layers")

    if Rb < 2*r_wire:
        Rb = 2*r_wire
    if Rp < R_outer_lyr + 2*(Rb + r_wire):
        Rp = R_outer_lyr + 2*(Rb + r_wire) + 1e-5
    if h < 2*(Rb + r_wire):
        h = 2*(Rb + r_wire) + 0.002

    l = np.pi*coil_span/Q * (Rp + R_inner_lyr)
    z = Rp - R_outer_lyr - 2*(Rb + r_wire)
    l_ew = 2*h + l + z + (Rb + r_wire)*(np.pi - 2)
    return l_ew, h, Rp, Rb

def end_wdg_hairpin_check(alpha, h, dmin, l_h, wire_w, tooth_wmin, h_conn):
    alpha = alpha*np.pi/180     # ensure alpha is in radians
    dmin_new = 0
    alpha_new = 0
    h_new = 0

    if dmin < 0.0015:
        dmin = 0.0015
    if h_conn < 0.005:
        h_conn = 0.005

    if alpha == 0 and h == 0: # end wdg parameters not set
        alpha_new = np.arcsin((dmin + wire_w)/(tooth_wmin + wire_w))
        h_new = np.tan(alpha_new)*l_h/2

    elif h > 0: # imposed end wdg height
        alpha_new = np.arctan(h/l_h)
        dmin_new = np.sin(alpha_new)*(tooth_wmin + wire_w) - wire_w
        if dmin_new < dmin: # imposed end wdg height is not feasible - calculate min end wdg parameters
            dmin_new = dmin
            alpha_new = np.arcsin((dmin_new + wire_w)/(tooth_wmin + wire_w))
            h_new = np.tan(alpha_new)*l_h/2

    elif alpha > 0: # imposed end wdg angle
        dmin_new = np.sin(alpha)*(tooth_wmin + wire_w) - wire_w
        if dmin_new < dmin: # imposed end wdg angle is not feasible - calculate min end wdg parameters
            dmin_new = dmin
            alpha_new = np.arcsin((dmin_new + wire_w)/(tooth_wmin + wire_w))
            h_new = np.tan(alpha_new)*l_h/2

    if dmin_new > dmin:
        dmin = dmin_new
    if alpha_new > alpha:
        alpha = alpha_new
    if h_new > h:
        h = h_new

    return h, alpha, dmin, h_conn

def end_wdg_length_hairpins(wire_h, wire_w, wire_th, wire_gap,
                                 layers, coil_pitch, Q, bore_diam, slot_w, slot_h,
                                 h_bent=0, h_welded=0, h_conn=0.005, alpha_bent=0, alpha_welded=0, dmin=0.0015): # needs to be validated
    '''return end wdg length of single pin for bent and welded side, average end wdg length,
     bent and welded side end wdg heights, bending angles and min distances between pins'''

    R_avg = bore_diam/2 + wire_th + layers/2*(wire_h + wire_gap) + wire_h/2
    l_h = R_avg*coil_pitch/Q*2*np.pi
    tooth_wmin = (bore_diam + 2*wire_th)*np.pi/Q - slot_w

    h_bent, alpha_bent, dmin, h_conn = end_wdg_hairpin_check(alpha_bent, h_bent, dmin, l_h, wire_w, tooth_wmin, h_conn)
    h_welded = h_welded - h_conn if h_welded - h_conn > 0 else 0
    h_welded, alpha_welded, dmin, h_conn = end_wdg_hairpin_check(alpha_welded, h_welded, dmin, l_h, wire_w, tooth_wmin, h_conn)

    l_bent = 2*(0.003 + wire_w/2*alpha_bent + wire_w*(np.pi/2 - alpha_bent) + np.sqrt((l_h/2)**2 + h_bent**2))
    l_welded = 2*(0.003 + wire_w/2*alpha_bent + wire_w*(np.pi/2 - alpha_bent) + np.sqrt((l_h/2)**2 + h_welded**2)) + h_conn
    l_ew = (l_bent + l_welded)/2
    h_welded = h_welded + h_conn
    return l_bent, l_welded, l_ew, h_bent, h_welded, h_conn, alpha_bent*180/np.pi, alpha_welded*180/np.pi, dmin


class Winding(object):
    def __init__(self, arg):
        """create winding either from bch winding section or winding dict()

        Arguments:
          arg: (object) femagtools.bch.Reader
            or dict Q: number of slots, m: number of phases, p: pole pairs, l: num layers, yd: coils span (slots)
       """
        if isinstance(arg, femagtools.bch.Reader):
            self.m = arg.machine['m']
            self.Q = arg.machine['Q']
            self.p = arg.machine['p']
            self.windings = arg.windings
        else:
            for k in arg.keys():
                setattr(self, k, arg[k])
        # balanced winding check
        if np.mod(self.Q, self.m*np.gcd(self.Q, self.p)):
            raise ValueError(
                f"Unbalanced winding: Q={self.Q}, p={self.p}, m={self.m}")
        self.q = self.Q/2/self.p/self.m  # number of coils per pole and phase

        if hasattr(self, 'windings'):
            # calculate coil width yd and num layers l
            taus = 360/self.Q
            try:
                ngen = arg.machine.get('qs_sim', 0)
            except AttributeError:
                ngen = 0
            if 'slots' in self.windings[1]:  # custom winding def
                for k in self.windings:
                    w = self.windings[k]
                    ngen = max(max(w['slots']), ngen)
                    w['dir'] = [1 if s > 0 else -1 for s in w['slots']]
                    w['PHI'] = [(2*abs(s)-1)*taus/2 for s in w['slots']]
                    w['R'] = [0 if l == 1 else 1 for l in w['layer']]

            self.yd = max(self.Q//self.p//2, 1)
            w = 1
            try:
                k = self.windings[w]['dir'].index(
                    -self.windings[w]['dir'][0])
                dphi = (self.windings[w]['PHI'][k] -
                        self.windings[w]['PHI'][0])
                self.yd = abs(round(dphi/taus))
            except ValueError:
                pass

            slots = [round((x-taus/2)/taus)
                     for x in self.windings[1]['PHI']]
            self.l = 1
            if len(slots) > ngen//self.m:
                self.l = 2
            return

        layers = 1
        if hasattr(self, 'l'):
            layers = self.l
        else:
            self.l = layers
        coilwidth = max(self.Q//self.p//2, 1)
        if hasattr(self, 'coilwidth'):   # obsolete, use yd instead
            coilwidth = self.coilwidth
        elif hasattr(self, 'yd'):
            coilwidth = self.yd

        self.yd = coilwidth
        if self.m % 2:  # odd number of phases
            q1, q2, Yk, Qb = q1q2yk(self.Q, self.p, self.m, self.l)
            j = 2 if layers == 1 else 1
            k1 = [(q1 + q2)*i for i in range(self.m)]
            k2 = (q1*(self.m+1) + q2*(self.m-1))//2
            pos = [[(j*Yk*(k + n)) % Qb
                    for n in range(q1)] for k in k1]
            neg = [[j*Yk*(k + n + k2) % Qb
                    for n in range(q2)] for k in k1]
        else:  # even number of phases
            if np.mod(self.Q, 4/self.l*self.m*np.gcd(self.Q, self.p)):
                raise ValueError(
                    f"Unbalanced winding: Q={self.Q}, p={self.p}, m={self.m}")
            t = np.gcd(self.Q, self.p)
            Qb = self.Q//t
            seq = np.array([i for j in range(Qb)
                            for i in range(Qb) if (j*self.p//t - i) % Qb == 0])
            if self.l == 1:
                seq = seq[::2]
            seq = np.array(seq).reshape((-1, 6))
            pos = seq[::2].T
            neg = seq[1::2].T

        if self.l > 1:
            slots = [sorted([(k, 1, 1)
                             for k in p] + [(k, -1, 1)
                                            for k in n])
                     for n, p in zip(neg, pos)]
            for i, p in enumerate(slots):
                slots[i] = sorted(slots[i] +
                                  [((k[0]+self.yd) % Qb, -k[1], 0)
                                   for k in slots[i]])
        else:
            if (coilwidth + 1) % 2:
                coilwidth += 1
            xneg = [sorted([s for s in n if s+1 % 2] +
                           [(s + coilwidth) % Qb for s in p if s+1 % 2])
                    for n, p in zip(neg, pos)]
            xpos = [sorted([s for s in p if s+1 % 2] +
                           [(s + coilwidth) % Qb for s in n if s+1 % 2])
                    for n, p in zip(neg, pos)]

            slots = [sorted([(k, 1, 1) for k in p] + [(k, -1, 1) for k in n])
                     for n, p in zip(xneg, xpos)]

        taus = 360/self.Q
        self.windings = {
            i+1:  dict(dir=[k[1] for k in s],
                       N=[1]*len(s), R=[k[2] for k in s],
                       PHI=[taus/2+k[0]*taus for k in s])
            for i, s in enumerate(slots)}

    def kw_order(self, n):
        """return winding factor harmonics"""
        if np.isscalar(n):
            if n == 0:
                return self.p
            g = np.arange(-n, n, 1)
            t = num_basic_windings(self.Q, self.p, self.l)
            return self.p + g * self.m*t
        return n

    def kwp(self, n=0):
        """pitch factor"""
        if np.asarray(n).any() and not np.isscalar(n):
            nue = n
        else:
            nue = self.kw_order(n)
        return np.sin(nue*self.yd*np.pi/self.Q)

    def kwd(self, n=0):
        """zone (distribution) factor"""
        q1, q2, Yk, Qb = q1q2yk(self.Q, self.p, self.m, self.l)
        if np.asarray(n).any() and not np.isscalar(n):
            nue = n
        else:
            nue = self.kw_order(n)
        if q1 == q2:  # integral slot winding
            q = self.Q/2/self.m/self.p
            nuep = nue/self.p
            return np.sin(nuep*np.pi/2/self.m)/q/np.sin(nuep*np.pi/2/self.m/q)
        k = 2 if self.l == 1 else 1
        a = nue*k*np.pi/self.Q*Yk
        t = self.Q//Qb
        return ((np.sin(a*q1) - np.cos(np.pi/t*Yk*nue)*np.sin(a*q2))
                /((q1+q2)*np.sin(a)))

    def kw(self, n=0):
        """return winding factor"""
        return np.abs(self.kwp(n) * self.kwd(n))

    def harmleakcoeff(self, n=4000):
        """return harmonic leakage coefficient
        Arguments:
        n: (int) maximum number of harmonics"""
        nue = self.p * (1 + np.arange(1-n, n)*2*self.m)
        kw1 = self.kwd()*self.kwp()
        kwn = self.kwd(nue)*self.kwp(nue)
        return np.sum((self.p*kwn/nue/kw1)**2) - 1

    def coils_per_phase(self):
        """return number of coils per phase"""
        return self.Q * self.l/2/self.m

    def turns_per_phase(self, n, g):
        """return number of turns per phase
        Arguments:
        n: (int) number of wires per slot side
        g: (int) number of parallel coil groups
        """
        return n*self.coils_per_phase()//g

    def inductance(self, nwires, g, da1, lfe, ag, sw=0):
        """return main inductance / phase
        Arguments:
        nwires: number of wires in slot side
        g: (int) number of parallel groups
        da1: (float) bore diameter / m
        lfe: (float) length of lamination /m
        ag: (float) length of airgap / m
        sw: (float) slot opening width / m
        """
        mue0 = 4*np.pi*1e-7
        # carter factor
        tauq = np.pi*da1/self.Q
        h = sw/ag
        xic = 2/np.pi*(h*np.arctan(h/2) - np.log(1+(h/2)**2))
        kc = tauq/(tauq-xic*ag)  # Carter factor
        de = kc * ag

        N1 = self.turns_per_phase(nwires, g)
        return 3*mue0*da1*lfe*((self.kw()*N1)**2 /
                               (np.pi*self.p**2*de))

    def sequence(self):
        """returns sequence of winding keys"""
        return list(zip(*sorted([(k, self.windings[k]['PHI'][0])
                                 for k in self.windings.keys()],
                                key=lambda wdg: wdg[1])))[0]

    def slots(self, key):
        """returns slot indexes of winding key"""
        taus = 360/self.Q
        t = num_basic_windings(self.Q, self.p, self.l)
        Qb = self.Q//t
        dim = self.l*Qb//self.m
        ngen = t
        qgen = Qb
        if dim > len(self.windings[key]['PHI']):
            qgen = qgen//2
            ngen = 2*t
        slots = [(round((x-taus/2)/taus) + qgen*n) % self.Q + 1
                 for n in range(ngen)
                 for x in self.windings[key]['PHI'][:dim]]
        return np.array(slots).reshape(t, -1)

    def axis(self, k=1):
        """returns axis angle of winding 1 in mechanical system"""
        return self.mmf(k=k)['alfa0']

    def currdist(self, k=1, phi=0):
        """return the current density of slots
        Arguments:
          k: (int) winding key (all if 0 or 'all')
          phi: (float) current angle (default 0)
        """
        if k == 0 or k == 'all':
            keys = self.windings.keys()
        elif np.isscalar(k):
            keys = [k]
        else:
            keys = k
        m = len(self.windings)
        t = num_basic_windings(self.Q, self.p, self.l)
        cdist = {k: np.zeros(self.Q//t) for k in keys}
        for z in self.zoneplan():
            if z:
                for j in keys:
                    a = np.zeros(self.Q//t)
                    curr = np.cos((j-1)*2*np.pi/m - phi)
                    for s in z[j-1]:
                        d = -1 if s < 0 else 1
                        a[abs(s)-1] = d*curr
                    cdist[j] += a
        return cdist

    def mmf(self, k=1, nmax=21, phi=0):
        """returns the dimensionless magnetomotive force
        (ampere-turns/turns/ampere) and winding angle of phase k (rad)
        Arguments:
        k: (int) winding key (all if 0 or 'all')
        nmax: (int) max order of harmonic (in electrical system)
        phi: (float) current angle (default 0)
        """
        cdist = self.currdist(k, phi)
        num_slots = len(cdist[list(cdist.keys())[0]])
        t = self.Q//num_slots
        clink = np.zeros(num_slots)
        for j in cdist:
            a = cdist[j]
            ap = np.zeros(len(a))
            l = 0
            v = 0
            for n in np.nonzero(a>0)[0]:
                ap[l:n] = v
                v += cdist[j][n]
                l = n
            ap[n:] = v
            an = np.zeros(len(a))
            l = 0
            v = 0
            for n in np.nonzero(a<0)[0]:
                an[l:n] = v
                v += cdist[j][n]
                l = n
            an[n:] = v
            clink += an + ap

        NY = 4096  # number of samples per slot
        y = [[c]*NY for c in (clink-np.mean(clink))]
        yy = np.tile(np.hstack(
            (y[-1][-NY//2:], np.ravel(y[:-1]), y[-1][:-NY//2])), t)
        yy /= np.max(yy)

        # calc spectrum
        pb = self.p//t
        N = len(yy)
        Y = np.fft.fft(yy)
        if self.q < 1:
            imax = pb
        else:
            imax = np.argmax(np.abs(Y[:N//2]))
        a = 2*np.abs(Y[imax])/N
        taus = 2*np.pi/self.Q
        freq = np.fft.fftfreq(N, d=taus/NY)
        T0 = np.abs(1/freq[imax])
        alfa0 = np.angle(Y[imax])
        pos_fft = np.linspace(0, 2*np.pi/t, 20*pb)
        D = (a*np.cos(2*np.pi*pos_fft/T0+alfa0))
        nue, mmf_nue = np.array(
            [(n, f) for n, f in zip(
                np.arange(0, nmax),
                2*np.abs(Y)/N) if f > 0]).T

        if alfa0 > 0:
            alfa0 -= 2*np.pi

        return dict(
            pos=(taus/NY*np.arange(NY*self.Q//t)).tolist(),
            mmf=yy[:NY*self.Q//t].tolist(),
            alfa0=-alfa0/self.p,
            nue=nue.tolist(),
            currdist={k: cdist[k].tolist() for k in cdist},
            mmf_nue=mmf_nue.tolist(),
            pos_fft=pos_fft.tolist(),
            mmf_fft=D.tolist())

    def zoneplan(self):
        taus = 360/self.Q
        dphi = 1e-3
        slots = {k: [s-1 for s in self.slots(k)[0]]
                 for k in self.windings}
        layers = 1
        avgr = 0
        maxr, minr = max(self.windings[1]['R']), min(self.windings[1]['R'])
        if maxr-minr > 1e-6:
            layers = 2
            avgr = (maxr+minr)/2

            def is_upper(r, phi):
                return r > avgr
        elif len(slots[1]) > len(set(slots[1])):
            layers = 2

            def is_upper(r, phi):
                return phi < -dphi
        else:
            def is_upper(r, phi):
                return True

        upper = [[s+1 for s, x, r in zip(
            slots[key],
            self.windings[key]['PHI'],
            self.windings[key]['R'])
            if is_upper(r, s*taus - (x-taus/2))]
            for key in self.windings]
        udirs = [[d for s, d, x, r in zip(
            slots[key],
            self.windings[key]['dir'],
            self.windings[key]['PHI'],
            self.windings[key]['R'])
            if is_upper(r, s*taus - (x-taus/2))]
            for key in self.windings]
        lower = []
        ldirs = []
        if layers > 1:
            lower = [[s+1 for s, x, r in zip(
                slots[key],
                self.windings[key]['PHI'],
                self.windings[key]['R'])
                if not is_upper(r, s*taus - (x-taus/2))]
                for key in self.windings]
            ldirs = [[d for s, d, x, r in zip(
                slots[key],
                self.windings[key]['dir'],
                self.windings[key]['PHI'],
                self.windings[key]['R'])
                if not is_upper(r, s*taus - (x-taus/2))]
                for key in self.windings]

        z = ([[int(d*s) for s, d in zip(u, ud)] for u, ud in zip(upper, udirs)],
             [[int(d*s) for s, d in zip(l, ld)] for l, ld in zip(lower, ldirs)])
        # complete if not  basic winding:
        Qb = self.Q//num_basic_windings(self.Q, self.p, self.l)

        if not np.asarray(upper).size or not np.asarray(lower).size:
            layers = 1
        if layers == 1 and z[1]:
            z = ([[int(d*s) for s, d in zip(l, ld)] for l, ld in zip(lower, ldirs)],
                 [[int(d*s) for s, d in zip(u, ud)] for u, ud in zip(upper, udirs)])

        if max([abs(n) for m in z[0] for n in m]) < Qb:
            return [[k + [-n+Qb//2 if n < 0 else -(n+Qb//2) for n in k]
                     for k in m] for m in z]
        return z

    def diagram(self) -> ET.Element:
        """return winding diagram as svg element"""
        coil_len = 25
        coil_height = 4
        dslot = 8
        arrow_head_length = 2
        arrow_head_width = 2
        strokewidth = [f"{w}px" for w in [0.25, 0.5]]

        z = self.zoneplan()
        xoff = 0
        if z[-1]:
            xoff = 0.75
        yd = dslot*self.yd
        mh = 2*coil_height/yd
        slots = sorted([abs(n) for m in z[0] for n in m])
        smax = slots[-1]*dslot
        ET.register_namespace("", "http://www.w3.org/2000/svg")
        svg = ET.Element("svg", dict(
            version="1.1", xmlns="http://www.w3.org/2000/svg",
            viewBox=f"0, -30, {slots[-1] * dslot + 15}, 40"))
        g = ET.SubElement(svg, "g", {"id": "teeth", "fill": "lightblue"})
        for n in slots:
            e = ET.SubElement(g, "rect", {
                "x": f"{n * dslot + dslot/4}",
                "y": f"{-coil_len + 1}",
                "width": f"{dslot/2}",
                "height": f"{coil_len - 2}"})

        g = ET.SubElement(svg, "g",
                          {"id": "labels",
                           "text-anchor": "middle",
                           "dominant-baseline": "middle",
                           "style": "font-size: 0.15em; font-family: sans-serif;"})
        for n in slots:
            t = ET.SubElement(g, "text", {
                "x": f"{n*dslot}",
                "y": f"{-coil_len / 2}"}).text = str(n)

        g = ET.SubElement(svg, "g", {"id": "coils",
                                     "fill": "none",
                                     "stroke-linejoin": "round",
                                     "stroke-linecap": "round"})

        direction = ['right', 'left']
        nl = 2 if z[1] else 1
        for i, layer in enumerate(z):
            b = -xoff if i else xoff
            w = i if self.yd > 1 else 0
            for m, mslots in enumerate(layer):
                for k in mslots:
                    slotpos = abs(k) * dslot + b
                    pc = [f"L {slotpos} {-coil_len//2+2} M {slotpos} {-coil_len//2-1}",
                          f"L {slotpos} {-coil_len}"]

                    if nl == 2:
                        if k > 0:
                            d = 0 if i == 0 else 1
                        else:
                            d = 1 if i == 1 else 0
                    else:
                        d = 0 if k > 0 else 1

                    if direction[d] == 'right':
                        # first layer, positive dir or neg. dir and 2-layers:
                        #   from right bottom
                        if slotpos + yd > smax+b:
                            dx = dslot if yd > dslot else yd/4
                            ph = [f"M {slotpos+yd//2-xoff+dx} {coil_height-mh*dx}",
                                  f"L {slotpos+yd//2-xoff} {coil_height} L {slotpos} 0"]
                            pt = [f"L {slotpos+yd//2-xoff} {-coil_len-coil_height}",
                                  f"L {slotpos+yd//2-xoff+dx} {-coil_len-coil_height+mh*dx}"]
                        else:
                            ph = [
                                f"M {slotpos+yd//2-xoff} {coil_height} L {slotpos} 0"]
                            pt = [
                                f"L {slotpos+yd//2-xoff} {-coil_len-coil_height}"]
                    else:
                        # from left bottom
                        if slotpos - yd < 0:  # and slotpos - yd > -3*dslot:
                            dx = dslot if yd > dslot else yd/4
                            ph = [f"M {slotpos-yd//2+xoff-dx} {coil_height-mh*dx}",
                                  f"L {slotpos-yd//2+xoff} {coil_height} L {slotpos} 0"]
                            pt = [f"L {slotpos-yd//2+xoff} {-coil_len-coil_height}",
                                  f"L {slotpos-yd//2+xoff-dx} {-coil_len-coil_height+mh*dx}"]
                        else:
                            ph = [
                                f"M {slotpos-yd//2+xoff} {coil_height} L {slotpos} 0"]
                            pt = [
                                f"L {slotpos-yd//2+xoff} {-coil_len-coil_height}"]
                    e = ET.SubElement(g, "path", {
                        "d": ' '.join(ph + pc + pt),
                        "stroke-width": strokewidth[w],
                        "stroke": coil_color[m]})

        for i, layer in enumerate(z):
            for m, mslots in enumerate(layer):
                for k in mslots:
                    x = abs(k) * dslot
                    if i:
                        x -= xoff
                    else:
                        x += xoff
                    if k > 0:
                        y = coil_len * .88
                        points = [
                            (x, -y),
                            (x - arrow_head_width / 2, -y + arrow_head_length),
                            (x + arrow_head_width / 2, -y + arrow_head_length)]
                    else:
                        y = coil_len * .12
                        points = [
                            (x, -y),
                            (x - arrow_head_width / 2, -y - arrow_head_length),
                            (x + arrow_head_width / 2, -y - arrow_head_length)]
                    ET.SubElement(svg, "polygon", {
                        "points": " ".join([f"{x},{y}" for (x, y) in points]),
                        "fill": f"{coil_color[m]}",
                        "stroke": "none"})

        return ET.tostring(svg, encoding='unicode')

    def write(self, name, workdir='.'):
        """creates WID file"""
        import pathlib
        with open(pathlib.Path(workdir) / (name + '.WID'), 'w') as fp:
            if 'slots' in self.windings[1]:
                inp = sorted([(k if s > 0 else -k, abs(s), l, n)
                              for k in self.windings
                              for s, l, n in zip(
                    self.windings[k]['slots'],
                    self.windings[k]['layer'],
                    self.windings[k]['N'])],
                    key=lambda x: (x[1], x[2]))
                fp.write('\n'.join([
                    f'Windings input data: {name}',
                    ' Number of coil sides:',
                    f'          {len(inp)}',
                    'Type of machine: 1 = Rot, 21 = Lin-x, 22 = Lin-y',
                    '           1',
                    'Index  w-keys     N-turns         Layer', '']))

                for i, t in enumerate(inp):
                    fp.write(f'{i+1:5d} {t[0]:8d} {t[3]:10.3f} {t[2]:10d}\n')
            else:
                inp = sorted([(k if d > 0 else -k, r, phi, n)
                              for k in self.windings
                              for d, r, phi, n in zip(
                    self.windings[k]['dir'],
                    self.windings[k]['R'],
                    self.windings[k]['PHI'],
                    self.windings[k]['N'])],
                    key=lambda x: (x[1], x[2]))
                fp.write('\n'.join([
                    f'Windings input data: {name}',
                    ' Number of coil sides:',
                    f'          {len(inp)}',
                    'Type of machine: 1 = Rot, 21 = Lin-x, 22 = Lin-y',
                    '           1',
                    'Index  w-keys     N-turns         R[mm]     PHI[deg]', '']))
                for i, t in enumerate(inp):
                    fp.write(
                        f'{i+1:5d} {t[0]:8d} {t[3]:10.3f} {t[1]:8.3f} {t[2]:8.3f}\n')
            fp.write('\n'.join([
                'Number of windings saved :',
                f'         {len(self.windings)}',
                'W-Key Coil-Current [A]   W-Types: (=1 :wire&cur)'
            ] + [
                f'   {k}          0.00000000       0.00000000                            1'
                for k in self.windings
            ] + ['   0', '']))


if __name__ == "__main__":
    import sys
    import matplotlib.pyplot as plt
    from xml.etree import ElementTree as ET

    if sys.argv[1:]:
        bch = femagtools.bch.read(sys.argv[1])
        wdgs = Winding(bch)
    else:
        testdata = [
            dict(Q=90, p=12, m=3, l=2, coilwidth=1),
            dict(Q=54, p=6, m=3, l=2, coilwidth=5),
            dict(Q=168, p=7, m=3, l=2, coilwidth=10)]

        wdgs = Winding(testdata[1])

    c = wdgs.mmf()
    # print('alfa0={0:6.3f}'.format(wdgs.axis()/np.pi*180))

    plt.title('Q={0}, p={1}, alfa0={2:6.3f}'.format(
        wdgs.Q, wdgs.p, c['alfa0']/np.pi*180))
    plt.plot(np.array(c['pos'])/np.pi*180, c['mmf'])
    plt.plot(np.array(c['pos_fft'])/np.pi*180, c['mmf_fft'])

    phi = [c['alfa0']/np.pi*180, c['alfa0']/np.pi*180]
    y = [min(c['mmf_fft']), 1.1*max(c['mmf_fft'])]
    plt.plot(phi, y, '--')
    plt.annotate("", xy=(phi[0], y[0]),
                 xytext=(0, y[0]), arrowprops=dict(arrowstyle="->"))

    plt.grid()
    plt.show()

    svg = wdgs.diagram()
    tree = ET.ElementTree(ET.fromstring(svg))
    tree.write('wind.svg')
    print('SVG file "wind.svg" created')
